/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/


/* bsColor.c */

#include "bsColor.h"

static int verbose = 0;

static int NORMALIZE=1;

static float NORMALFLOOR=0.75;
static float BGUPDATERATE=0.1;
static int IGNOREBLACK=0;

#include "eigen.h"

void bsColorIgonreBlack() {
  printf("ignore black called \n");
  IGNOREBLACK=1;
}

float bsColorGetUpdateBG() {
  return BGUPDATERATE;
}

void bsColorSetUpdateBG(float newvalue) {
  BGUPDATERATE = newvalue;
}

float bsColorGetNormalFloor() {
  return NORMALFLOOR;
}

void bsColorSetNormalFloor(float newvalue) {
  NORMALFLOOR = newvalue;
}

bsColorGetNormalize() {
	return NORMALIZE;
}
void bsColorSetNormalize(int newvalue) {
	NORMALIZE = newvalue;
}


void
bsColorSetVerbose( int val )
{
  verbose = val;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * printMatrix
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void
printMatrix(Matrix A)
{
  int c1, c2;
  
  fprintf(stderr,"\nNew Matrix:\n");
  for (c1 = 0; c1 < A->height; c1++)
    {
      fprintf(stderr,"\n");
      for (c2 = 0; c2 < A->width; c2++)
	fprintf(stderr,"%8.5f ",A->data[c1][c2]);
    }
  
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorDiscrim
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

ftype
bsColorDiscrim(ftype *rawcolor, color_class *cclass, int norm_type)
{
  extern int NORMALIZE;  
  ftype dist[3], tmp_t[3];
  register ftype a, result, ch0;
  register ftype *meanp = cclass->mean;
  register ftype **matp;
  ftype norm;
  ftype color[3];

#ifdef DIGITIZEYUV
	  a = rawcolor[0]/(meanp[0]);
#else
	  a = (rawcolor[0]+rawcolor[1]+rawcolor[2]) /
	    (meanp[0]+meanp[1]+meanp[2]);
#endif
  
  switch(norm_type)
    {
    case BG_NORMALIZE:
      if (!NORMALIZE || a<NORMALFLOOR || a>=0.98) 
	break;
      
    case ALL_NORMALIZE:
      
      norm = rawcolor[0];
#ifdef DIGITIZEYUV
      color[0] = rawcolor[1];
      color[1] = rawcolor[2];
#else
      color[0] = rawcolor[1] / norm; 
      color[1] = rawcolor[2] / norm;
#endif

      meanp = cclass->norm_mean;

      diff2(color,meanp,dist);
      matp = cclass->norm_inv_covar->data;
      diag_transpose_mul_mat2(dist,matp,tmp_t);
      result = - dot2(tmp_t,dist) - cclass->norm_ln_det_covar;
      return result;

    case NO_NORMALIZE:
      break;
      
    }

  diff3(rawcolor,meanp,dist);
  matp = cclass->inv_covar->data;
  diag_transpose_mul_mat3(dist,matp,tmp_t);
  result = - dot3(tmp_t,dist) - cclass->ln_det_covar;
  
  return result;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXYDiscrim
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

ftype
bsXYDiscrim(int x, int y, color_class *cclass, int zoom)
{
  ftype result;

  if (!cclass->fgp) {
    fprintf(stderr,"bsXYDiscrim called on a background pixel!!!\n");
    result = 0.0;
  } else {
    ftype data[2], dist[2], tmp_t[2];
    register ftype *meanp = cclass->fg->xy_mean;
    register ftype **matp = cclass->fg->xy_inv_covar->data;
    
    data[0] = x; data[1] = y;
    
    diff2(data,meanp,dist);
    dist[0] /= zoom; dist[1] /= zoom;
    transpose_mul_mat2(dist,matp,tmp_t);
    result = - dot2(tmp_t,dist) - cclass->fg->xy_ln_det_covar;
  }
  return result;
}

int bsXYBoundingBox(int x, int y, color_class *cclass, int zoom)
{
  int result;

  if (!cclass->fgp) {
    fprintf(stderr,"bsXYBoundingBox called on a background pixel!!!\n");
    result = 0;
  } else {
    ftype data[2];
    ftype dist[2];

    register ftype *mean   = cclass->fg->xy_mean;
    register ftype *bb   = cclass->fg->bb;
    register ftype **covar = cclass->fg->xy_inv_covar->data;
    
    data[0] = (ftype) x;
    data[1] = (ftype) y;
    
    dist[0] = bsABS(data[0] - mean[0]) / (ftype) zoom; 
    dist[1] = bsABS(data[1] - mean[1]) / (ftype) zoom;
  
    result = ((dist[0] < (10.0 * bb[0])) &&
	      (dist[1] < (10.0 * bb[1])));
  }

  return (result);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorComputeForegroundClass
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void updatepixel(ivePixel *ivepixel,
			color_class **fore,
			int num_classes,
			int class, 
			int x, int y)
{
  register int c,k,n;
  register ftype p, p0;
  ftype xy[2];
  
  xy[0] = x;
  xy[1] = y;
  
  fore[class]->fg->num_points += 1;

  for (c = 0; c < 3; c++) {
    p = ivePixelChannel( ivepixel, c );
    fore[class]->fg->mean_acc[c] += p;
  }

  p0 = ivePixelChannel( ivepixel, 0 );
  for (c = 0; c < 2; c++) {
#ifdef DIGITIZEYUV
    p = ivePixelChannel(ivepixel,c+1);
#else
    p = ivePixelChannel(ivepixel,c+1)/p0;
#endif
    fore[class]->fg->norm_mean_acc[c] += p;
  }

  for (c = 0; c < 3; c++) {
    p = ivePixelChannel( ivepixel, c );
    fore[class]->fg->covar_acc->data[c][c] += p*p;
  }

  p0 = ivePixelChannel( ivepixel, 0 );
  for (c = 0; c < 2; c++) {
#ifdef DIGITIZEYUV
    p = ivePixelChannel(ivepixel,c+1);
#else
    p = ivePixelChannel(ivepixel,c+1)/p0;
#endif
    fore[class]->fg->norm_covar_acc->data[c][c] += p*p;
  }

  for (c = 0; c < 2; c++) {
    fore[class]->fg->xy_mean_acc[c] += xy[c];
  }

  for (c = 0; c < 2; c++) {
    p = xy[c];
    for (k = 0; k < 2; k++) {
      fore[class]->fg->xy_covar_acc->data[c][k] += p * xy[k];
    }
  }
}

void bsColorInitClasses(color_class **fore,
			int num_classes)
{
  int class;
  int c,k;

  /* clear mean and covar for each class */
  
  for (class=0; class<num_classes; class++)
    {
      fore[class]->fg->last_num_points=fore[class]->fg->num_points;
      fore[class]->fg->num_points=0;

      for (c = 0; c < 3; c++) {
	fore[class]->fg->mean_acc[c] = 0.0;
      }
      
      for (c = 0; c < 2; c++) {
	fore[class]->fg->norm_mean_acc[c] = 0.0;
      }
      
      for (c = 0; c < 3; c++) {
	fore[class]->fg->covar_acc->data[c][c] = 0.0;
      }
      for (c = 0; c < 2; c++) {
	fore[class]->fg->norm_covar_acc->data[c][c] = 0.0;
      }

      for (c = 0; c < 2; c++) {
	fore[class]->fg->xy_mean_acc[c] = 0.0;
      }
  
      for (c = 0; c < 2; c++) {
	for (k = 0; k < 2; k++) {
	  fore[class]->fg->xy_covar_acc->data[c][k] = 0.0;
	}
      }
    }
}


int bsColorAddExample(int classnum,
		      int xpos,
		      int ypos,
		      ivePixel *ivep,
		      color_class **fore,
		      int num_classes)
{
  /* for each pixel... add it to the appropriate class */
  /* if it's foreground.  always add to class 1 */
	
  switch (classnum)
    {
    case 0:
      return;
      break;

    case 1:
      updatepixel(ivep,fore,num_classes,0,xpos,ypos);
      break;
      
    default:
      updatepixel(ivep,fore,num_classes,classnum-1,xpos,ypos);
      updatepixel(ivep,fore,num_classes,0,xpos,ypos);
      break;
    }
}
  

void bsColorBlendInPriors(color_class *fore)
{
  int c,k;
  register int   copy = (fore->singular); 
  register ftype cmpriorweight = (copy ? 1.0 : fore->fg->cmpriorweight);
  register ftype cvpriorweight = (copy ? 1.0 : fore->fg->cvpriorweight);
  register ftype xympriorweight= (copy ? 1.0 : fore->fg->xympriorweight);
  register ftype xyvpriorweight= (copy ? 1.0 : fore->fg->xyvpriorweight);
  register ftype cmdataweight  = 1.0 - cmpriorweight;
  register ftype cvdataweight  = 1.0 - cvpriorweight;
  register ftype xymdataweight = 1.0 - xympriorweight;
  register ftype xyvdataweight = 1.0 - xyvpriorweight;

  fore->singular = 0;

  if((fore->fg->class_num != pfHEAD) && 
     (fore->fg->class_num != pfRIGHTHAND) && 
     (fore->fg->class_num != pfLEFTHAND))
    for (c = 0; c < 3; c++) {
      fore->mean[c] =
	(fore->mean[c] * cmdataweight +
	 fore->fg->prior_mean[c] * cmpriorweight);
    }

  for (c = 0; c < 2; c++) {
    fore->norm_mean[c] =
      (fore->norm_mean[c] * cmdataweight +
       fore->fg->prior_norm_mean[c] * cmpriorweight);
  }
  for (c = 0; c < 2; c++) {
    fore->fg->xy_mean[c] =
      (fore->fg->xy_mean[c] * xymdataweight +
       fore->fg->prior_xy_mean[c] * xympriorweight);
  }
  for (c = 0; c < 3; c++) {
    fore->inv_covar->data[c][c] =
      ((fore->inv_covar->data[c][c] * 
	cvdataweight) +
       (fore->fg->prior_inv_covar->data[c][c] * 
	cvpriorweight));
  }
  for (c = 0; c < 2; c++) {
    fore->norm_inv_covar->data[c][c] =
      ((fore->norm_inv_covar->data[c][c] * 
	cvdataweight) +
       (fore->fg->prior_norm_inv_covar->data[c][c] * 
	cvpriorweight));
  }
  for (c = 0; c < 2; c++) {
    for (k = 0; k < 2; k++) {
      fore->fg->xy_inv_covar->data[c][k] =
	((fore->fg->xy_inv_covar->data[c][k] * 
	  xyvdataweight) +
	 (fore->fg->prior_xy_inv_covar->data[c][k] * 
	  xyvpriorweight));
    }
  }
  
  
  fore->ln_det_covar = ((fore->ln_det_covar * 
				cvdataweight) +
			       (fore->fg->prior_ln_det_covar * 
				cvpriorweight));
  
  fore->norm_ln_det_covar = ((fore->norm_ln_det_covar * 
				     cvdataweight) +
				    (fore->fg->prior_norm_ln_det_covar * 
				     cvpriorweight));
  
  fore->fg->xy_ln_det_covar = ((fore->fg->xy_ln_det_covar * 
				  xyvdataweight) +
				 (fore->fg->prior_xy_ln_det_covar * 
				  xyvpriorweight));
  
  fore->fg->bb[0] = ((fore->fg->bb[0]  * 
			xyvdataweight) +
		       (fore->fg->prior_bb[0]  * 
			xyvpriorweight));
  fore->fg->bb[1] = ((fore->fg->bb[1] * 
			xyvdataweight) +
		       (fore->fg->prior_bb[1] * 
			xyvpriorweight));
  
}

void bsColorSetFromPriors(color_class *fore)
{
  int c,k;
  
  for (c = 0; c < 3; c++) {
    fore->mean[c] = fore->fg->prior_mean[c];
  }
  for (c = 0; c < 2; c++) {
    fore->norm_mean[c] = fore->fg->prior_norm_mean[c];
  }
  for (c = 0; c < 2; c++) {
    fore->fg->xy_mean[c] = fore->fg->prior_xy_mean[c];
  }
  for (c = 0; c < 3; c++) {
    fore->inv_covar->data[c][c] = fore->fg->prior_inv_covar->data[c][c];
  }
  for (c = 0; c < 2; c++) {
    fore->norm_inv_covar->data[c][c] = 
      fore->fg->prior_norm_inv_covar->data[c][c];
  }
  for (c = 0; c < 2; c++) {
    for (k = 0; k < 2; k++) {
      fore->fg->xy_inv_covar->data[c][k] = 
	fore->fg->prior_xy_inv_covar->data[c][k];
    }
  }
  
  
  fore->ln_det_covar = fore->fg->prior_ln_det_covar;
  
  fore->norm_ln_det_covar = fore->fg->prior_norm_ln_det_covar;
  
  fore->fg->xy_ln_det_covar = fore->fg->prior_xy_ln_det_covar;
  
  fore->fg->bb[0] = fore->fg->prior_bb[0];

  fore->fg->bb[1] = fore->fg->prior_bb[1];  
}

static ftype diag_MatrixDeterminant(Matrix A)
{
  register ftype result =1.0;
  register int i;
  
  for (i=0;i<A->height;i++)
    result *= A->data[i][i];

  return result;
}

static ftype MatrixDeterminant_2(Matrix A)
{
  return(A->data[0][0]*A->data[1][1] - 
	 A->data[0][1]*A->data[1][0]); 
}

void diag_MatrixInvert(Matrix A)
{
  register int i;
  
  for (i=0;i<A->height;i++)
    A->data[i][i] = 1.0 / A->data[i][i];
}

void MatrixInvert_2(Matrix A)
{
  if ( MatrixDeterminant_2(A) != 0.)
    {
      Real b = A->data[0][0];
      Real c = A->data[0][1];
      Real d = A->data[1][0];
      Real e = A->data[1][1];
      
      Real cd = c*d;
      Real ib = (1./b);
      Real ie = (1./e);
      
      Real x = 1. / (b - ie*cd);
      Real u = 1. / (e - ib*cd);
      Real y = -ib*c*u;
      Real z = -ie*d*x;
      
      A->data[0][0] = x;
      A->data[0][1] = y;
      A->data[1][0] = z;
      A->data[1][1] = u;
    }
}


  
int
bsColorComputeForegroundClasses( iveBitmap *inMap,
				iveImage *frame,
				color_class **fore,
				int num_classes)
{
  static ivePixel ivepixel;
  static ftype *num = NULL;
  int i,j,c,k,pixel,n;
  register ftype p, xy[2];
  int class;
  
  int m = frame -> pixelCount;
  int w = frame -> width;
  
  /* compute Sum(x), Sum(x^T x) */
  
  /* calculate stats for each class */
  
  for (class=0; class<num_classes; class++)
    {
      /* compute E[x], E[x^T x] */
      
      if (fore[class]->fg->num_points > 3 )
	{
	  for (c = 0; c < 3; c++) {
	    fore[class]->mean[c] = 
	      fore[class]->fg->mean_acc[c] / 
		fore[class]->fg->num_points;

	  }
	  for (c = 0; c < 2; c++) {
	    fore[class]->norm_mean[c] = 
	      fore[class]->fg->norm_mean_acc[c] / 
		fore[class]->fg->num_points;
	  }

	  for (c = 0; c < 3; c++) {
	    fore[class]->inv_covar->data[c][c] =
	      fore[class]->fg->covar_acc->data[c][c] /
		fore[class]->fg->num_points;
	  }   
	  for (c = 0; c < 2; c++) {
	    fore[class]->norm_inv_covar->data[c][c] = 
	      fore[class]->fg->norm_covar_acc->data[c][c] / 
		fore[class]->fg->num_points;
	  }

	  for (c = 0; c < 2; c++) {
	    fore[class]->fg->xy_mean[c] = 
	      fore[class]->fg->xy_mean_acc[c] / 
		fore[class]->fg->num_points;
	  }

	  for (c = 0; c < 2; c++) {
	    for (k = 0; k < 2; k++) {
	      fore[class]->fg->xy_inv_covar->data[c][k] = 
		fore[class]->fg->xy_covar_acc->data[c][k] /
		  fore[class]->fg->num_points;
	    }
	  }
	  
	  
	  /* compute E[(x-m)^T (x-m)] = E[x^T x] - m m^T] */
	  
	  for (c = 0; c < 3; c++) {
	    p = fore[class]->mean[c];
	    fore[class]->inv_covar->data[c][c] -= p*p;
	  }
	  for (c = 0; c < 2; c++) {
	    p = fore[class]->norm_mean[c];
	    fore[class]->norm_inv_covar->data[c][c] -= p*p;
	  }

	  for (c = 0; c < 2; c++) {
	    for (k = 0; k < 2; k++) {
	      fore[class]->fg->xy_inv_covar->data[c][k] 
		-= fore[class]->fg->xy_mean[c] * 
		  fore[class]->fg->xy_mean[k];
	    }
	  }
	  
	  for (c = 0; c < 2; c++) 
	    {
	      fore[class]->fg->bb[c]=fore[class]->fg->xy_inv_covar->data[c][c];
	      fore[class]->fg->bb[c] = bsSQRT(fore[class]->fg->bb[c]);  
	    }

	  /* compute inverse and log determinant of E[(x-m)^T (x-m)] */
	  /* compute ln of det later because you want that value now */
	  
	  fore[class]->ln_det_covar = 
	    diag_MatrixDeterminant(fore[class]->inv_covar);
	  fore[class]->norm_ln_det_covar = 
	    diag_MatrixDeterminant(fore[class]->norm_inv_covar);
	  
	  fore[class]->fg->xy_ln_det_covar = 
	    MatrixDeterminant_2(fore[class]->fg->xy_inv_covar);
	  
	  
	  /* compute principle axes of blob from covariance matrix */

	  eigen(fore[class]->fg->xy_inv_covar->data,
		fore[class]->fg->xyL,
		fore[class]->fg->xyS);
	  fore[class]->fg->xyL[0] = sqrt(fore[class]->fg->xyL[0]);
	  fore[class]->fg->xyL[1] = sqrt(fore[class]->fg->xyL[1]);

	  /* Deal with Non-invertible matrices */
	  
	  fore[class]->singular = 1;
	  
	  if (fore[class]->ln_det_covar != 0.0) 
	    {
	      diag_MatrixInvert(fore[class]->inv_covar);
	      if (fore[class]->norm_ln_det_covar != 0.0) 
		{
		  diag_MatrixInvert(fore[class]->norm_inv_covar);
		  fore[class]->ln_det_covar = 
		    bsLOG(bsABS(fore[class]->ln_det_covar));
		  fore[class]->norm_ln_det_covar = 
		    bsLOG(bsABS(fore[class]->norm_ln_det_covar));
		  
		  fore[class]->singular = 0;
		}
	    }
	  
	  
	  if (fore[class]->fg->xy_ln_det_covar != 0.0) 
	    {
	      MatrixInvert_2(fore[class]->fg->xy_inv_covar);
	      fore[class]->fg->xy_ln_det_covar = 
		bsLOG(bsABS(fore[class]->fg->xy_ln_det_covar));
	      
	      fore[class]->singular |= 0;
	    }
	}
      else
	{
	  fore[class]->singular = 1;
	}	  
      
      if (fore[class]->fg->prior_set)  
	{
	  bsColorBlendInPriors(fore[class]);
	}
    }
  return 1;
}


static int bsColorPrintClass(color_class **fore, int class)
{
      printf("class %d is marked %s\n",class,
	     (fore[class]->singular ? "singular" : "valid"));
      printf("\t color mean %f %f %f\n",
	     (float) fore[class]->mean[0],
	     (float) fore[class]->mean[1],
	     (float) fore[class]->mean[2]);
      printf("\t color covar %g %g %g\n",
	     (float) fore[class]->inv_covar->data[0][0],
	     (float) fore[class]->inv_covar->data[0][1],
	     (float) fore[class]->inv_covar->data[0][2]);
      printf("\t             %g %g %g\n",
	     (float) fore[class]->inv_covar->data[1][0],
	     (float) fore[class]->inv_covar->data[1][1],
	     (float) fore[class]->inv_covar->data[1][2]);
      printf("\t             %g %g %g\n",
	     (float) fore[class]->inv_covar->data[2][0],
	     (float) fore[class]->inv_covar->data[2][1],
	     (float) fore[class]->inv_covar->data[2][2]);
      printf("\t ln_det_covar %f\n",
	     (float) fore[class]->ln_det_covar);
      printf("\t norm2 mean %f %f\n",
	     (float) fore[class]->norm_mean[0],
	     (float) fore[class]->norm_mean[1]);
      printf("\t norm2 covar %g %g\n",
	     (float) fore[class]->norm_inv_covar->data[0][0],
	     (float) fore[class]->norm_inv_covar->data[0][1]);
      printf("\t             %g %g\n",
	     (float) fore[class]->norm_inv_covar->data[1][0],
	     (float) fore[class]->norm_inv_covar->data[1][1]);
      printf("\t norm_ln_det %f\n",
	     (float) fore[class]->norm_ln_det_covar);
      printf("\t xy    mean %f %f\n",
	     (float) fore[class]->fg->xy_mean[0],
	     (float) fore[class]->fg->xy_mean[1]);
      printf("\t xy    covar %g %g\n",
	     (float) fore[class]->fg->xy_inv_covar->data[0][0],
	     (float) fore[class]->fg->xy_inv_covar->data[0][1]);
      printf("\t             %g %g\n\n",
	     (float) fore[class]->fg->xy_inv_covar->data[1][0],
	     (float) fore[class]->fg->xy_inv_covar->data[1][1]);
      printf("\t xy_ln_det %f\n",
	     (float) fore[class]->fg->xy_ln_det_covar);
}	     

int bsColorSetClassColor(color_class *fore, pfColors pfc)
{
  switch(pfc){

  case pfFLESH:
#ifndef DIGITIZEYUV
    fore->fg->prior_mean[0] = 167.3392;
    fore->fg->prior_mean[1] = 125.0676;
    fore->fg->prior_mean[2] = 116.2794;
    fore->fg->prior_inv_covar->data[0][0] =  0.0046;
    fore->fg->prior_inv_covar->data[0][1] = -0.0035;
    fore->fg->prior_inv_covar->data[0][2] = -0.0022;
    fore->fg->prior_inv_covar->data[1][0] = -0.0035;
    fore->fg->prior_inv_covar->data[1][1] =  0.1271;
    fore->fg->prior_inv_covar->data[1][2] = -0.1269;
    fore->fg->prior_inv_covar->data[2][0] = -0.0022;
    fore->fg->prior_inv_covar->data[2][1] = -0.1269;
    fore->fg->prior_inv_covar->data[2][2] =  0.1350;
    fore->fg->prior_ln_det_covar = 14.2607;
    
    fore->fg->prior_norm_mean[0] = 0.7482;
    fore->fg->prior_norm_mean[1] = 0.6944;
    fore->fg->prior_norm_inv_covar->data[0][0] =  3.3721e+03;
    fore->fg->prior_norm_inv_covar->data[0][1] = -3.3068e+03;
    fore->fg->prior_norm_inv_covar->data[1][0] = -3.3068e+03;
    fore->fg->prior_norm_inv_covar->data[1][1] =  3.5530e+03;
    fore->fg->prior_norm_ln_det_covar = -13.8604;
#else
    fore->fg->prior_mean[0] = 1.1210e+02;
    fore->fg->prior_mean[1] = 1.1897e+02;
    fore->fg->prior_mean[2] = 1.4304e+02;
    fore->fg->prior_inv_covar->data[0][0] =  1.0817e-03;
    fore->fg->prior_inv_covar->data[0][1] =  4.0954e-03;
    fore->fg->prior_inv_covar->data[0][2] = -4.2127e-03;
    fore->fg->prior_inv_covar->data[1][0] =  4.0954e-03;
    fore->fg->prior_inv_covar->data[1][1] =  1.7511e-01;
    fore->fg->prior_inv_covar->data[1][2] =  4.2571e-02;
    fore->fg->prior_inv_covar->data[2][0] = -4.2127e-03;
    fore->fg->prior_inv_covar->data[2][1] =  4.2571e-02;
    fore->fg->prior_inv_covar->data[2][2] =  8.3670e-02;
    fore->fg->prior_ln_det_covar = 1.1748e+01;
    
    fore->fg->prior_norm_mean[0] = 1.1897e+02;
    fore->fg->prior_norm_mean[1] = 1.4304e+02;
    fore->fg->prior_norm_inv_covar->data[0][0] =  1.5960e-00;
    fore->fg->prior_norm_inv_covar->data[0][1] =  5.8521e-01;
    fore->fg->prior_norm_inv_covar->data[1][0] =  5.8521e-01;
    fore->fg->prior_norm_inv_covar->data[1][1] =  6.7263e-01;
    fore->fg->prior_norm_ln_det_covar = 4.9184e+00;
#endif
    break;
    
  case pfBROWN:
#ifndef DIGITIZEYUV
    fore->fg->prior_mean[0] = 1.1375e+02;
    fore->fg->prior_mean[1] = 1.1488e+02;
    fore->fg->prior_mean[2] = 1.0821e+02;

    fore->fg->prior_inv_covar->data[0][0] =  2.2707e-02;
    fore->fg->prior_inv_covar->data[0][1] = -3.1389e-02;
    fore->fg->prior_inv_covar->data[0][2] =  9.6011e-03;
    fore->fg->prior_inv_covar->data[1][0] = -3.1389e-02;
    fore->fg->prior_inv_covar->data[1][1] =  6.2528e-02;
    fore->fg->prior_inv_covar->data[1][2] = -3.3824e-02;
    fore->fg->prior_inv_covar->data[2][0] =  9.6011e-03;
    fore->fg->prior_inv_covar->data[2][1] = -3.3824e-02;
    fore->fg->prior_inv_covar->data[2][2] =  2.6413e-02;
    fore->fg->prior_ln_det_covar =  1.5920e+01;
    
    fore->fg->prior_norm_mean[0] = 1.0337e+00;
    fore->fg->prior_norm_mean[1] = 1.0011e+00;
    fore->fg->prior_norm_inv_covar->data[0][0] =  1.8318e+02;
    fore->fg->prior_norm_inv_covar->data[0][1] = -8.5223e+01;
    fore->fg->prior_norm_inv_covar->data[1][0] = -8.5223e+01;
    fore->fg->prior_norm_inv_covar->data[1][1] =  5.4902e+01;
    fore->fg->prior_norm_ln_det_covar = -7.9352e+00;
#else
    fore->fg->prior_mean[0] = 1.1374e+02;
    fore->fg->prior_mean[1] = 1.2524e+02;
    fore->fg->prior_mean[2] = 1.2787e+02;

    fore->fg->prior_inv_covar->data[0][0] =  5.8286e-04;
    fore->fg->prior_inv_covar->data[0][1] =  6.5214e-03;
    fore->fg->prior_inv_covar->data[0][2] =  4.4571e-03;
    fore->fg->prior_inv_covar->data[1][0] =  6.5214e-03;
    fore->fg->prior_inv_covar->data[1][1] =  1.7095e-01;
    fore->fg->prior_inv_covar->data[1][2] =  1.2613e-01;
    fore->fg->prior_inv_covar->data[2][0] =  4.4571e-03;
    fore->fg->prior_inv_covar->data[2][1] =  1.2613e-01;
    fore->fg->prior_inv_covar->data[2][2] =  1.8059e-01;
    fore->fg->prior_ln_det_covar =  1.2210e+01;
    
    fore->fg->prior_norm_mean[0] =  1.2524e+02;
    fore->fg->prior_norm_mean[1] =  1.2787e+02;
    fore->fg->prior_norm_inv_covar->data[0][0] =  9.7988e-02;
    fore->fg->prior_norm_inv_covar->data[0][1] =  7.6257e-02;
    fore->fg->prior_norm_inv_covar->data[1][0] =  7.6257e-02;
    fore->fg->prior_norm_inv_covar->data[1][1] =  1.4651e-01;
    fore->fg->prior_norm_ln_det_covar = 4.7629e+00;
#endif
    break;
    
    
  case pfGREY:
#ifndef DIGITIZEYUV
    fore->fg->prior_mean[0] = 1.3245e+02;
    fore->fg->prior_mean[1] = 1.3346e+02;
    fore->fg->prior_mean[2] = 1.3279e+02;

    fore->fg->prior_inv_covar->data[0][0] =  1.6478e-01;
    fore->fg->prior_inv_covar->data[0][1] = -8.8287e-02;
    fore->fg->prior_inv_covar->data[0][2] = -7.5562e-02;
    fore->fg->prior_inv_covar->data[1][0] = -8.8287e-02;
    fore->fg->prior_inv_covar->data[1][1] =  1.7639e-01;
    fore->fg->prior_inv_covar->data[1][2] = -8.7783e-02;
    fore->fg->prior_inv_covar->data[2][0] = -7.5562e-02;
    fore->fg->prior_inv_covar->data[2][1] = -8.7783e-02;
    fore->fg->prior_inv_covar->data[2][2] =  1.6704e-01;
    fore->fg->prior_ln_det_covar = 9.1638e+00;
    
    fore->fg->prior_norm_mean[0] = 1.0686e+00;
    fore->fg->prior_norm_mean[1] = 1.0779e+00;
    fore->fg->prior_norm_inv_covar->data[0][0] =  2.3699e+02;
    fore->fg->prior_norm_inv_covar->data[0][1] = -1.9269e+02;
    fore->fg->prior_norm_inv_covar->data[1][0] = -1.9269e+02;
    fore->fg->prior_norm_inv_covar->data[1][1] =  2.3163e+02;
    fore->fg->prior_norm_ln_det_covar = -9.7849e+00;
#else
    fore->fg->prior_mean[0] = 1.2719e+02;
    fore->fg->prior_mean[1] = 1.2766e+02;
    fore->fg->prior_mean[2] = 1.2784e+02;

    fore->fg->prior_inv_covar->data[0][0] =  7.1406e-03;
    fore->fg->prior_inv_covar->data[0][1] =  6.4219e-04;
    fore->fg->prior_inv_covar->data[0][2] =  4.9953e-03;
    fore->fg->prior_inv_covar->data[1][0] =  6.4219e-04;
    fore->fg->prior_inv_covar->data[1][1] =  8.2171e-01;
    fore->fg->prior_inv_covar->data[1][2] = -1.9697e-02;
    fore->fg->prior_inv_covar->data[2][0] =  4.9953e-03;
    fore->fg->prior_inv_covar->data[2][1] = -1.9697e-02;
    fore->fg->prior_inv_covar->data[2][2] =  7.8254e-01;
    fore->fg->prior_ln_det_covar = 5.3887e+00;
    
    fore->fg->prior_norm_mean[0] = 1.2784e+02;
    fore->fg->prior_norm_mean[1] = 1.2766e+02;
    fore->fg->prior_norm_inv_covar->data[0][0] =  8.2165e-01;
    fore->fg->prior_norm_inv_covar->data[0][1] = -2.0146e-02;
    fore->fg->prior_norm_inv_covar->data[1][0] = -2.0146e-02;
    fore->fg->prior_norm_inv_covar->data[1][1] =  7.7905e-01;
    fore->fg->prior_norm_ln_det_covar = 4.4676e-01;
#endif
    break;
    
    
  case pfZERO:
    fore->fg->prior_mean[0] = 0.0;
    fore->fg->prior_mean[1] = 0.0;
    fore->fg->prior_mean[2] = 0.0;
    fore->fg->prior_inv_covar->data[0][0] = 0.0;
    fore->fg->prior_inv_covar->data[0][1] = 0.0;
    fore->fg->prior_inv_covar->data[0][2] = 0.0;
    fore->fg->prior_inv_covar->data[1][0] = 0.0;
    fore->fg->prior_inv_covar->data[1][1] = 0.0;
    fore->fg->prior_inv_covar->data[1][2] = 0.0;
    fore->fg->prior_inv_covar->data[2][0] = 0.0;
    fore->fg->prior_inv_covar->data[2][1] = 0.0;
    fore->fg->prior_inv_covar->data[2][2] = 0.0;
    fore->fg->prior_ln_det_covar =  1.0e3;

    fore->fg->prior_norm_inv_covar->data[0][0] = 0.0;
    fore->fg->prior_norm_inv_covar->data[0][1] = 0.0;
    fore->fg->prior_norm_inv_covar->data[1][0] = 0.0;
    fore->fg->prior_norm_inv_covar->data[1][1] = 0.0;
    fore->fg->prior_norm_ln_det_covar = 1.0e3;
    break;
    
  case pfRED:
    /*(r > 1.5 * b && r > 1.5 * g && r > 20.0) */
#ifndef DIGITIZEYUV
    fore->fg->prior_mean[0] = 1.9470e+02;
    fore->fg->prior_mean[1] = 6.6510e+01;
    fore->fg->prior_mean[2] = 6.5324e+01;
    fore->fg->prior_inv_covar->data[0][0] =  1.5525e-02;
    fore->fg->prior_inv_covar->data[0][1] = -6.1210e-03;
    fore->fg->prior_inv_covar->data[0][2] = -5.8863e-03;
    fore->fg->prior_inv_covar->data[1][0] = -6.1210e-03;
    fore->fg->prior_inv_covar->data[1][1] =  1.6958e-02;
    fore->fg->prior_inv_covar->data[1][2] =  7.6803e-04;
    fore->fg->prior_inv_covar->data[2][0] = -5.8863e-03;
    fore->fg->prior_inv_covar->data[2][1] =  7.6803e-04;
    fore->fg->prior_inv_covar->data[2][2] =  1.6661e-02;
    fore->fg->prior_ln_det_covar =  1.2646e+01;
    
    fore->fg->prior_norm_mean[0] = 3.4160e-01;
    fore->fg->prior_norm_mean[1] = 3.3591e-01;
    fore->fg->prior_norm_inv_covar->data[0][0] =  6.9305e+02;
    fore->fg->prior_norm_inv_covar->data[0][1] =  3.5158e+01;
    fore->fg->prior_norm_inv_covar->data[1][0] =  3.5158e+01;
    fore->fg->prior_norm_inv_covar->data[1][1] =  6.5616e+02;
    fore->fg->prior_norm_ln_det_covar = -1.3025e+01;
#else
    fore->fg->prior_mean[0] = 1.0511e+02;
    fore->fg->prior_mean[1] = 1.0896e+02;
    fore->fg->prior_mean[2] = 1.8379e+02;
    fore->fg->prior_inv_covar->data[0][0] =  3.2942e-02;
    fore->fg->prior_inv_covar->data[0][1] =  2.1815e-02;
    fore->fg->prior_inv_covar->data[0][2] = -6.5378e-03;
    fore->fg->prior_inv_covar->data[1][0] =  2.1815e-02;
    fore->fg->prior_inv_covar->data[1][1] =  7.3643e-02 ;
    fore->fg->prior_inv_covar->data[1][2] = -1.0368e-02;
    fore->fg->prior_inv_covar->data[2][0] = -6.5378e-03;
    fore->fg->prior_inv_covar->data[2][1] = -1.0368e-02;
    fore->fg->prior_inv_covar->data[2][2] =  6.1463e-02;
    fore->fg->prior_ln_det_covar = 9.0608e+00;
    
    fore->fg->prior_norm_mean[0] = 1.0896e+02;
    fore->fg->prior_norm_mean[1] = 1.8379e+02;
    fore->fg->prior_norm_inv_covar->data[0][0] =  5.9197e-02;
    fore->fg->prior_norm_inv_covar->data[0][1] = -6.0387e-03;
    fore->fg->prior_norm_inv_covar->data[1][0] = -6.0387e-03;
    fore->fg->prior_norm_inv_covar->data[1][1] =  6.0165e-02;
    fore->fg->prior_norm_ln_det_covar =  5.6478e+00;
#endif
    break;

  case pfGREEN:
    /* (g - r > 30 && g - b > 30) */
#ifndef DIGITIZEYUV
    fore->fg->prior_mean[0] = 8.6180e+01;
    fore->fg->prior_mean[1] = 1.9905e+02;
    fore->fg->prior_mean[2] = 8.3386e+01;
    fore->fg->prior_inv_covar->data[0][0] =  9.8188e-03;
    fore->fg->prior_inv_covar->data[0][1] = -4.9201e-03;
    fore->fg->prior_inv_covar->data[0][2] =  2.2798e-04;
    fore->fg->prior_inv_covar->data[1][0] = -4.9201e-03;
    fore->fg->prior_inv_covar->data[1][1] =  1.8060e-02;
    fore->fg->prior_inv_covar->data[1][2] = -5.2599e-03;
    fore->fg->prior_inv_covar->data[2][0] =  2.2798e-04;
    fore->fg->prior_inv_covar->data[2][1] = -5.2599e-03;
    fore->fg->prior_inv_covar->data[2][2] =  1.0062e-02;
    fore->fg->prior_ln_det_covar = 1.3569e+01;
    
    fore->fg->prior_norm_mean[0] = 5.5369e+00;
    fore->fg->prior_norm_mean[1] = 2.1970e+00;
    fore->fg->prior_norm_inv_covar->data[0][0] =  5.7541e-01;
    fore->fg->prior_norm_inv_covar->data[0][1] = -1.0582e+00;
    fore->fg->prior_norm_inv_covar->data[1][0] = -1.0582e+00;
    fore->fg->prior_norm_inv_covar->data[1][1] =  3.0985e+00;
    fore->fg->prior_norm_ln_det_covar = 4.1074e-01;
#else
    fore->fg->prior_mean[0] = 1.4716e+02;
    fore->fg->prior_mean[1] = 9.3649e+01;
    fore->fg->prior_mean[2] = 8.6080e+01;
    fore->fg->prior_inv_covar->data[0][0] =  2.4832e-02;
    fore->fg->prior_inv_covar->data[0][1] =  7.0936e-03;
    fore->fg->prior_inv_covar->data[0][2] =  1.8435e-03;
    fore->fg->prior_inv_covar->data[1][0] =  7.0936e-03;
    fore->fg->prior_inv_covar->data[1][1] =  5.2324e-02;
    fore->fg->prior_inv_covar->data[1][2] =  1.6259e-02;
    fore->fg->prior_inv_covar->data[2][0] =  1.8435e-03;
    fore->fg->prior_inv_covar->data[2][1] =  1.6259e-02;
    fore->fg->prior_inv_covar->data[2][2] =  5.0842e-02;
    fore->fg->prior_ln_det_covar = 9.7692e+0;
    
    fore->fg->prior_norm_mean[0] = 9.3649e+01;
    fore->fg->prior_norm_mean[1] = 8.6080e+01;
    fore->fg->prior_norm_inv_covar->data[0][0] =  5.0298e-02;
    fore->fg->prior_norm_inv_covar->data[0][1] =  1.5732e-02;
    fore->fg->prior_norm_inv_covar->data[1][0] =  1.5732e-02;
    fore->fg->prior_norm_inv_covar->data[1][1] =  5.0705e-02;
    fore->fg->prior_norm_ln_det_covar = 6.0736e+00;
#endif
    break;
  }    
}	     

int bsColorSetClassSize(color_class *fore, int flag)
{
  switch (flag) {
  case XYWIDE:
    fore->fg->prior_xy_inv_covar->data[0][0] =  3.310e-03;
    fore->fg->prior_xy_inv_covar->data[0][1] =  1.096e-03;
    fore->fg->prior_xy_inv_covar->data[1][0] =  1.096e-03;
    fore->fg->prior_xy_inv_covar->data[1][1] =  5.473e-03;
    fore->fg->prior_xy_ln_det_covar = 16.8388;
    fore->fg->bb[0] = 17.0;
    fore->fg->bb[1] = 17.0;
    break;

  case XYMEDIUM:
    fore->fg->prior_xy_inv_covar->data[0][0] =  2.50000e-01;
    fore->fg->prior_xy_inv_covar->data[0][1] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][0] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][1] =  2.50000e-01;
    fore->fg->prior_xy_ln_det_covar = 4.00000e+00;
    fore->fg->prior_bb[0] = 2.0;
    fore->fg->prior_bb[1] = 2.0;
    break;

  case XYNARROW:
    fore->fg->prior_xy_inv_covar->data[0][0] =  1.00000e+00;
    fore->fg->prior_xy_inv_covar->data[0][1] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][0] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][1] =  1.00000e+00;
    fore->fg->prior_xy_ln_det_covar = 0.00000e+00;
    fore->fg->prior_bb[0] = 1.0;
    fore->fg->prior_bb[1] = 1.0;
    break;

  case XYTINY:
    fore->fg->prior_xy_inv_covar->data[0][0] =  4.0000e+00;
    fore->fg->prior_xy_inv_covar->data[0][1] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][0] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][1] =  4.0000e+00;
    fore->fg->prior_xy_ln_det_covar = -1.3862944e+00;
    fore->fg->prior_bb[0] = 0.5;
    fore->fg->prior_bb[1] = 0.5;
    break;
  }
}	     

int bsColorSetClassPosition(color_class *fore, ftype x, ftype y)
{
  fore->fg->xy_mean[0] = x;
  fore->fg->xy_mean[1] = y;
  fore->fg->prior_xy_mean[0] = x;
  fore->fg->prior_xy_mean[1] = y;
}	     


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorComputeBackgroundClassesBatch
 *
 \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <sys/time.h>

void
bsColorComputeBackgroundClassesBatch( iveImage **frames,
				     int frameCount,  
				     color_class **background )

{
  static ivePixel ivepixel;
  int pixelCount = frames[0]->pixelCount;
  int w = frames[0]->width;
  int i,j,c,k,pixel,rankzero=0;

  register ftype p=0, p0=0;

  struct timeval start;
  if( verbose ) 
    gettimeofday(&start);

  /* fprintf(stderr,"background pixelCount: %d\n",pixelCount); */

  /* clear mean and covar */

  if( verbose )
    {
      fprintf(stderr,"Computing background class now\n");
    }
  for (i=0; i<pixelCount; i++) {
    for (c=0; c<3; c++) {
      background[i]->mean[c] = 0.0;
      background[i]->inv_covar->data[c][c] = 0.0;
    }
    for (c=0; c<2; c++) {
      background[i]->norm_mean[c] = 0.0;
      background[i]->norm_inv_covar->data[c][c] = 0.0;
    }
    
  }
  

  {
    int CURRENT_PIXEL=0;
    long *pixel_count_per_stack=malloc(sizeof(long)*pixelCount);
    
    /* compute Sum(x), Sum(x^T x) */
    
    /* I thought that rearranging the loops so j was on the outside   */
    /* would speed things up... but that's not the case.              */
    /* ij ordering is empirically about twice as fast as ji ordering. */
    /* go fig....                                                     */

    for (i=0; i<pixelCount; i++) {
      pixel_count_per_stack[i]=0;

      for ( j = 0; j < frameCount; j++ ) {
	
	iveImageGetPixelI( frames[j], i, &ivepixel );
	
	if(IGNOREBLACK) {
	  for ( c=0; c<3; c++ )
	    p += ivePixelChannel( &ivepixel, c );
	  CURRENT_PIXEL = p;
	}
	else
	  CURRENT_PIXEL = 1;
	
	if(CURRENT_PIXEL > 0) {
	  pixel_count_per_stack[i]++;
	  for ( c=0; c<3; c++ ) {
	    p = ivePixelChannel( &ivepixel, c );
	    background[i]->mean[c] += p;
	    background[i]->inv_covar->data[c][c] += p*p;
	  }
	  p0 = ivePixelChannel( &ivepixel, 0 );
	  for ( c=0; c<2; c++ ) {
#ifdef DIGITIZEYUV
	    p = ivePixelChannel(&ivepixel,c+1);
#else
	    p = ivePixelChannel(&ivepixel,c+1)/p0;
#endif
	    background[i]->norm_mean[c] += p;
	    background[i]->norm_inv_covar->data[c][c] += p*p;
	  }
	}
      }
    }
    
    /* compute E[x], E[x^T x] */
    for (i=0; i<pixelCount; i++) {
      for (c=0; c<3; c++) {
	if(pixel_count_per_stack[i]!=0) {
	  background[i]->mean[c] /= (ftype) pixel_count_per_stack[i];
	  background[i]->inv_covar->data[c][c] /= 
	    (ftype) pixel_count_per_stack[i];
	}
      }
      for (c=0; c<2; c++) {
	if(pixel_count_per_stack[i]!=0) {
	  background[i]->norm_mean[c] /= (ftype) pixel_count_per_stack[i];
	  background[i]->norm_inv_covar->data[c][c] /= 
	    (ftype) pixel_count_per_stack[i];
	}
      }
    }

    free(pixel_count_per_stack);
  }

  /* compute E[(x-m)^T (x-m)] = E[x^T x] - m m^T  */
  
  for (i=0; i<pixelCount; i++) {
    for (c=0; c<3; c++) {
      p = background[i]->mean[c];
      background[i]->inv_covar->data[c][c] -= p*p;
    }
    for (c=0; c<2; c++) {
      p = background[i]->norm_mean[c];
	background[i]->norm_inv_covar->data[c][c] -= p*p;
    }
  }
  
  /* compute inverse and log determinant of E[(x-m)^T (x-m)] */
  
  for (i=0; i<pixelCount; i++) {

    background[i]->ln_det_covar = 
      diag_MatrixDeterminant(background[i]->inv_covar);
    background[i]->norm_ln_det_covar = 
      diag_MatrixDeterminant(background[i]->norm_inv_covar);

    if (background[i]->ln_det_covar == 0.0) {
      background[i]->inv_covar->data[0][0] = 4.0;
      background[i]->inv_covar->data[1][0] = 0.0;
      background[i]->inv_covar->data[2][0] = 0.0;
      background[i]->inv_covar->data[0][1] = 0.0;
      background[i]->inv_covar->data[1][1] = 4.0;
      background[i]->inv_covar->data[2][1] = 0.0;
      background[i]->inv_covar->data[0][2] = 0.0;
      background[i]->inv_covar->data[1][2] = 0.0;
      background[i]->inv_covar->data[2][2] = 4.0;
      background[i]->ln_det_covar = -4.1589;
      rankzero++; 
    } else {
      diag_MatrixInvert(background[i]->inv_covar);
      background[i]->ln_det_covar = 
	bsLOG(bsABS(background[i]->ln_det_covar));
    }


    if (background[i]->norm_ln_det_covar == 0.0) {
      background[i]->norm_inv_covar->data[0][0] =  3.2127;
      background[i]->norm_inv_covar->data[1][0] = -3.1999;
      background[i]->norm_inv_covar->data[0][1] = -3.1999;
      background[i]->norm_inv_covar->data[1][1] =  3.1875;
      background[i]->norm_ln_det_covar = 7.1390;
    } else {
      diag_MatrixInvert(background[i]->norm_inv_covar);
      background[i]->norm_ln_det_covar = 
	bsLOG(bsABS(background[i]->norm_ln_det_covar));
    }
  }
  
  if (rankzero>pixelCount/10) {
    fprintf(stderr, "%d of %d background classes are null!\n", 
	    rankzero, pixelCount);
    fprintf(stderr, "Default covariences set for these classes.\n");
  }
  if( verbose ) 
    fprintf(stderr,"Computed Background Classes ");

  if( verbose ) 
    {
      struct timeval end;
      double time_sec;
      gettimeofday( &end );
      time_sec = ((end.tv_usec - start.tv_usec)/1000000.0 + 
		  (end.tv_sec - start.tv_sec));
      fprintf ( stderr, " took %.6f sec", time_sec );
    }

  if( verbose ) 
    fprintf(stderr,"\n");
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorInstallChromakey
 *
 \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void
bsColorInstallChromakey(int r, int g, int b,
			color_class **background,
			int width, int height)
			
{
  register ivePixel ivepixel;
  register int pixelCount = width*height;
  int i,j,c,k,pixel,rankzero=0;
  register ftype p, p0;

  ivePixelChannel( &ivepixel, 0 ) = r;
  ivePixelChannel( &ivepixel, 1 ) = g;
  ivePixelChannel( &ivepixel, 2 ) = b;

  /* fprintf(stderr,"background pixelCount: %d\n",pixelCount); */

  /* clear mean and covar */

  if( verbose )
    {
      fprintf(stderr,"Setting Chromakey background class now\n");
    }

  for (i=0; i<pixelCount; i++) {

    background[i]->mean[0] = (float) r;
    background[i]->mean[1] = (float) g;
    background[i]->mean[2] = (float) b;

    if (r == 0)
      {
	background[i]->norm_mean[0] = (float) g;
	background[i]->norm_mean[1] = (float) b;
      }
    else 
      {
	background[i]->norm_mean[0] = (float) g / (float) r;
	background[i]->norm_mean[1] = (float) b / (float) r;
      }

    background[i]->inv_covar->data[0][0] = 4.0;
    background[i]->inv_covar->data[1][0] = 0.0;
    background[i]->inv_covar->data[2][0] = 0.0;
    background[i]->inv_covar->data[0][1] = 0.0;
    background[i]->inv_covar->data[1][1] = 4.0;
    background[i]->inv_covar->data[2][1] = 0.0;
    background[i]->inv_covar->data[0][2] = 0.0;
    background[i]->inv_covar->data[1][2] = 0.0;
    background[i]->inv_covar->data[2][2] = 4.0;
    background[i]->ln_det_covar = -4.1589;
    
    background[i]->norm_inv_covar->data[0][0] =  3.2127;
    background[i]->norm_inv_covar->data[1][0] = -3.1999;
    background[i]->norm_inv_covar->data[0][1] = -3.1999;
    background[i]->norm_inv_covar->data[1][1] =  3.1875;
    background[i]->norm_ln_det_covar = 7.1390;
  }
  
  if( verbose ) 
    fprintf(stderr,"Computed Background Classes\n");

}

void
bsColorUpdateBackgroundPixel(ivePixel *ivep, color_class *bg)
{
  register int c;
  register ftype p,p0;
  
  for ( c=0; c<3; c++ ) 
    {
      p = ivePixelChannel( ivep, c );
      bg->mean[c] = ((   BGUPDATERATE    * p) + 
		     ((1.0-BGUPDATERATE) * bg->mean[c]));
    }

  p0 = ivePixelChannel( ivep, 0 );
  for ( c=0; c<2; c++ ) 
    {
#ifdef DIGITIZEYUV
      p = ivePixelChannel(ivep,c+1);
#else
      p = ivePixelChannel(ivep,c+1)/p0;
#endif
      bg->norm_mean[c] = ((  BGUPDATERATE     * p) + 
			  ((1.0-BGUPDATERATE) * bg->norm_mean[c]));
    }
}

#define MIN( a, b ) ( ( (a) < (b) ) ? (a) : (b) )
#define MAX( a, b ) ( ( (a) > (b) ) ? (a) : (b) )
#define LIMIT( x, low, high ) MAX( (low), MIN( (x), (high) ) )

bsColorRgbToYuv(float *color)
{
  float r = color[0];
  float g = color[1];
  float b = color[2];
  
  float y = ( 0.257*r + 0.504*g + 0.098*b) +  16.0;
  float u = (-0.148*r - 0.291*g + 0.439*b) + 128.0;
  float v = ( 0.439*r - 0.368*g - 0.072*b) + 128.0;
  
  y = LIMIT (y , 0.0, 255.0);
  u = LIMIT (u , 0.0, 255.0);
  v = LIMIT (v , 0.0, 255.0);
  
  color[0] = y;
  color[1] = u;
  color[2] = v;
}

bsColorYuvToRgb(float *color)
{
  float y = color[0] - 16.0;
  float u = color[1] - 128.0;
  float v = color[2] - 128.0;

  float r = 1.164*y - 0.003*u + 1.596*v;
  float g = 1.161*y - 0.397*u - 0.814*v;
  float b = 1.162*y + 2.014*u - 0.001*v;
  
  r = LIMIT (r , 0.0, 255.0);
  g = LIMIT (g , 0.0, 255.0);
  b = LIMIT (b , 0.0, 255.0);
  
  color[0] = r;
  color[1] = g;
  color[2] = b;
}

#ifdef DIGITIZEYUV              /* if we're using YUV then identity */
void bsColorToYuv(float *color) { }
void bsColorYuvTo(float *color) { }
#else				/* if we're using YUV, then do it */
void bsColorToYuv(float *color) { bsColorRgbToYuv(color); }
void bsColorYuvTo(float *color) { bsColorYuvToRgb(color); }
#endif

#ifdef DIGITIZEYUV              /* if we're using YUV then do it */
void bsColorToRgb(float *color) { bsColorYuvToRgb(color); }
void bsColorRgbTo(float *color) { bsColorRgbToYuv(color); }
#else                           /* if we're using RGB, then identity */
void bsColorToRgb(float *color) { }
void bsColorRgbTo(float *color) { }
#endif
