/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

/* bsColor.h */

/* RCS: $Header: /v/projects/pfinder/distribution/src2/../../src/RCS/bsColor.h,v 9.0 1997/03/08 15:56:57 cwren Exp $ */

#ifndef _bsColor_H_
#define _bsColor_H_

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "matrix.h"
#include "iveBitmap.h"
#include "iveImage.h"

typedef double ftype;		/* make sure these 4 lines  */
#define bsEXP exp		/* are consistent!!!!!!!!!! */
#define bsLOG log		/* they should be: */
#define bsSQRT sqrt		/* ftype FUNC(ftype) */
/* #define bsABS abs */		/* ftype FUNC(ftype) */ 
#define bsABS fabs

#define diff2(b,c,r) (r)[0]=(b)[0]-(c)[0];(r)[1]=(b)[1]-(c)[1];
#define diff3(b,c,r) (r)[0]=(b)[0]-(c)[0];(r)[1]=(b)[1]-(c)[1];(r)[2]=(b)[2]-(c)[2];

#define diag_transpose_dot_mat2(c,k,n) ((c)[(n)]*(k)[(n)][(n)])
#define diag_transpose_dot_mat3(c,k,n) ((c)[(n)]*(k)[(n)][(n)])

#define diag_transpose_mul_mat2(c,k,r) \
  r[0]=diag_transpose_dot_mat2(c,k,0);\
  r[1]=diag_transpose_dot_mat2(c,k,1);
#define diag_transpose_mul_mat3(c,k,r) \
  r[0]=diag_transpose_dot_mat3(c,k,0);\
  r[1]=diag_transpose_dot_mat3(c,k,1);\
  r[2]=diag_transpose_dot_mat3(c,k,2);

#define transpose_dot_mat2(c,k,n) ((c)[0]*(k)[(n)][0]+(c)[1]*(k)[(n)][1])
#define transpose_dot_mat3(c,k,n) ((c)[0]*(k)[(n)][0]+(c)[1]*(k)[(n)][1]+(c)[2]*(k)[(n)][2])

#define transpose_mul_mat2(c,k,r) \
  r[0]=transpose_dot_mat2(c,k,0);\
  r[1]=transpose_dot_mat2(c,k,1);
#define transpose_mul_mat3(c,k,r) \
  r[0]=transpose_dot_mat3(c,k,0);\
  r[1]=transpose_dot_mat3(c,k,1);\
  r[2]=transpose_dot_mat3(c,k,2);

#define dot2(b,c) ((b)[0]*(c)[0] + (b)[1]*(c)[1])
#define dot3(b,c) ((b)[0]*(c)[0] + (b)[1]*(c)[1] + (b)[2]*(c)[2])


typedef struct fg_class_s {
  /* foreground only */ 
  ftype xy_mean[2];	/* screen xy */
  Matrix xy_inv_covar;	
  ftype xy_ln_det_covar;
  ftype bb[2];
  float xyL[2];			/* eigen values */
  float xyS[2][2];		/* eigen vectors */

  /* foreground accumulators */ 
  ftype mean_acc[3];	
  Matrix covar_acc;
  ftype norm_mean_acc[3];
  Matrix norm_covar_acc;
  ftype xy_mean_acc[3];
  Matrix xy_covar_acc;

  /* prior weights */ 
  float xympriorweight;
  float xyvpriorweight;
  float cmpriorweight;
  float cvpriorweight;

  /* color/xy weights */ 
  float cxyweight;

  /* priors */
  ftype prior_mean[3];		
  Matrix prior_inv_covar;
  ftype prior_ln_det_covar;
  ftype prior_norm_mean[2];	
  Matrix prior_norm_inv_covar;
  ftype prior_norm_ln_det_covar;
  ftype prior_xy_mean[2];	
  ftype prior_xy_ln_det_covar;
  Matrix prior_xy_inv_covar;
  ftype prior_bb[2];

  /* foreground flags */
  int prior_set;
  int prior_singular;
  int num_points;
  int last_num_points;
  
  int class_num;
} fg_class;

typedef struct color_class_s {
  /* used by both fg and bg */
  int singular;
  int fgp;

  ftype mean[3];		/* 3d color */
  Matrix inv_covar;
  ftype ln_det_covar;
  
  ftype norm_mean[2];		/* 2d color */
  Matrix norm_inv_covar;
  ftype norm_ln_det_covar;

  /* foreground only */ 
  fg_class *fg;

} color_class;

#define pvec3(color) printf("{%f, %f, %f}\n", (color)[0], (color)[1], (color)[2]);
#define pvec2(color) printf("{%f, %f}\n", (color)[0], (color)[1]);
#define pmat(mat) printMatrix(mat);

typedef enum  { NO_NORMALIZE=0, ALL_NORMALIZE, BG_NORMALIZE } pfNormTypes;

typedef enum  { pfFLESH, pfGREEN, pfRED, pfBROWN, pfZERO, pfGREY} pfColors;

typedef enum  { XYTINY=0, XYNARROW, XYMEDIUM, XYWIDE } pfXYCovarTypes;

typedef enum  { pfGLOBAL=0, pfRIGHTHAND, pfLEFTHAND,
		  pfHEAD, pfRIGHTFOOT, pfLEFTFOOT,
		  pfCHEST, pfTHIGHS, pfWAIST         } pfClassDesig;

#include "bsColor.proto"

#endif /* _bsColor_H_ */
