/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

/* bsSegment.c */

/* RCS: $Header: /v/projects/pfinder/distribution/src2/../../src/RCS/bsSegment.c,v 9.0 1997/03/08 15:56:57 cwren Exp $ */

#include "bsSegment.h"
#include "bsSegmentP.h" /* is this really necessary? */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * GLOBAL VARIABLES
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static ZoomGrowFunction zoomfunction = Init; 

static int RECURSIVE      = 0;
static int SMOOTH 	  = 1;
static int ZOOM_GROW      = 1;
static int BICHSEL        = 0;
static int CHAIN_ITERATE  = 0;
static int DOUBLE_ITERATE = 0;
float bsSegmentTHRESH  = 150.0;


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * GLOBAL FUNCTIONS
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

float bsSegmentGetThresh()
{
  return (bsSegmentTHRESH);
}

float bsSegmentGetThreshMax()
{
  return (300.0);
}

void bsSegmentSetThresh(float newvalue)
{
  bsSegmentTHRESH = newvalue;
}


int bsSegmentGetZoomGrow()
{
  return(ZOOM_GROW);
}

void bsSegmentSetZoomGrow(int newvalue)
{
  ZOOM_GROW = newvalue;
  SetZoomGrowFunction();
}


int bsSegmentGetRecursive()
{
  return(RECURSIVE);
}

void bsSegmentSetRecursive(int newvalue)
{
  RECURSIVE = newvalue;
  SetZoomGrowFunction();
}

int bsSegmentGetSmooth() 
{
  return SMOOTH;
}
void bsSegmentSetSmooth(int newvalue) {
  SMOOTH = newvalue;
}

int bsSegmentGetBichsel()
{
  return(BICHSEL);
}

void bsSegmentSetBichsel(int newvalue)
{
  BICHSEL = newvalue;
  SetZoomGrowFunction();
}

int bsSegmentGetDoubleIterate()
{
  return(DOUBLE_ITERATE);
}

void bsSegmentSetDoubleIterate(int newvalue)
{
  DOUBLE_ITERATE = newvalue;
  SetZoomGrowFunction();
}

int bsSegmentGetChainIterate()
{
  return(CHAIN_ITERATE);
}

void bsSegmentSetChainIterate(int newvalue)
{
  CHAIN_ITERATE = newvalue;
  SetZoomGrowFunction();
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Init
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static int *visitedbulk = NULL;
static int **visited = NULL;
static PixelClassStruct *memo = NULL;

static int Init(iveImage *im,
		 iveBitmap *outMap,
		 iveBitmap *inMap,
		 int zoom,
		 Point *centroid )
{
  static int init = 0;

  if (im == (iveImage *) NULL)
    {
      return init;
    }
  else
    {
      register int width = im->width;
      register int height = im->height;
      int i;

      init = 1;

      visitedbulk = (int *) malloc(sizeof(int)*width*height);
      bzero(visitedbulk, sizeof(int)*width*height);

      visited = (int **) malloc(sizeof(int *)*width);
      for (i=0;i<width;i++)
	{
	  visited[i] = &visitedbulk[i*height];
	}
      memo = (PixelClassStruct *) 
	malloc(sizeof(PixelClassStruct)*width*height);
      bzero(memo, sizeof(PixelClassStruct)*width*height);

      SetZoomGrowFunction();
      return ((*zoomfunction)( im, outMap, inMap, zoom, centroid));
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsSegmentFindSilhouette
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int 
bsSegmentFindSilhouette(iveImage *im,
			iveBitmap *outMap,
			iveBitmap *inMap,
			int zoom,
			Point *centroid )
{

  register int width = im->width;
  register int height = im->height;

  iveBitmapZero( outMap );

  (*zoomfunction)( im, outMap, inMap, zoom, centroid);

  bzero(memo, sizeof(PixelClassStruct)*width*height);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * LOCAL FUNCTIONS
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * SetZoomGrowFunction
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


static void SetZoomGrowFunction()
{
  if (!Init((iveImage *) NULL, (iveBitmap *) NULL,
	   (iveBitmap *) NULL, 0, (Point *) NULL))
    {
      zoomfunction = Init;
    }
  else
    {
      if (ZOOM_GROW)
	{
	  if (RECURSIVE)
	    zoomfunction = ZoomGrow;
	  else
	    zoomfunction = Grow;

	}
      else
	{
	  zoomfunction = Raster;
	}
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * PixelClass
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* return boolean decision about pixel being n background   */
/* use neighborhood information */

static PixelClassStruct *CheckCoarserGrid(int x, int y, 
					  int zoomin, int w, int h)
{
  static PixelClassStruct pix;
  register int   zoom  = zoomin * 2;
  register int   gridx = x - (x % zoom);
  register int   gridy = y - (y % zoom);
  register int   index = w*(gridy)+gridx;
  register int  gindex = index;
  register int   zeros = 0;
  register int    ones = 0;
  
  
  if (!memo[index].valid || !memo[index].c ) 
    {
      zeros++;
    } 
  else 
    {
      gindex = index;
      ones++;
    }
  
  if (gridx+zoom < w) 
    {
      index = w*(gridy)+(gridx+zoom);
      if(!memo[index].valid || !memo[index].c) 
	{
	  zeros++;
	} 
      else
	{
	  gindex = index;
	  ones++;
	}

      if (gridy+zoom < h) 
	{
	  index = w*(gridy+zoom)+(gridx);
	  if(!memo[index].valid || !memo[index].c) 
	    {
	      zeros++;
	    } 
	  else 
	    {
	      gindex = index;
	      ones++;
	    }
	  
	  index = w*(gridy+zoom)+(gridx+zoom);
	  if(!memo[index].valid || !memo[index].c) 
	    {
	      zeros++;
	    } 
	  else
	    {
	      gindex = index;
	      ones++;
	    }
	}
    }
  
  if (zeros > ones) 
    return (NULL);

  pix.c    = memo[gindex].c;
  pix.dist = memo[gindex].dist;
  
  return (&pix);
}
  

float getonepix(iveImage *im, int zoom, int w, int h, int x, int y) 
{
  register index = w*y+x;
  
  if (x<0 || y<0 || x>=w || y>=h)
    {
      return (0);
    }

  if (!memo[index].valid)
    {
      PixelClassStruct *pix;
      
      pix = bsClassifyPixel(im, x, y, zoom);

      memcpy((void *)(&memo[index]),
	     (void *)pix,
	     sizeof(PixelClassStruct));
      memo[index].valid = 1;
    }

  return ((memo[index].dist));
}


static int PixelClass( iveImage *im,
		      int zoom,
		      int x,
		      int y )
{
  float thresh = -bsSegmentTHRESH; 
  int w = im->width;
  int h = im->height;
  int index = y*w + x;
  int ret;
  float Pneighbor;
  float PnMax, PnSum;
  float result;
  extern SMOOTH;
  
  result = getonepix(im, zoom, w, h, x, y); 
  if (SMOOTH) 
    {
      
      Pneighbor =  getonepix(im, zoom, w, h, x+zoom, y); 
      PnSum = PnMax = Pneighbor; 
      
      Pneighbor =  getonepix(im, zoom, w, h, x-zoom, y);  
      PnSum += Pneighbor;
      PnMax = (Pneighbor > PnMax ? Pneighbor : PnMax);
      
      Pneighbor =  getonepix(im, zoom, w, h, x, y+zoom); 
      PnSum += Pneighbor;
      PnMax = (Pneighbor > PnMax ? Pneighbor : PnMax);
      
      Pneighbor =  getonepix(im, zoom, w, h, x, y-zoom);  
      PnSum += Pneighbor;
      PnMax = (Pneighbor > PnMax ? Pneighbor : PnMax);
      
      if (BICHSEL) {
	result += PnMax; 
      } else {
	result = result * 0.5 + PnSum * (0.125 * 4) ;  
      }
    }
  
  if (result <= thresh)
    {
      register int c = memo[w*y+x].c;
      register int result = (c ? c : 1);
      bsXPaintBit(result,x,y);
      return (result);
    }
  else
    {
      return (0);
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * PixelSimple()
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* return a boolean decision about pixel being in background */

static int
PixelSimple( iveImage *im, int x, int y, int zoom)
{
  PixelClassStruct *pix = bsClassifyPixel( im, x, y, zoom) ;
  return (pix->c);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * GetEdgePoint
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int 
GetEdgePoint(iveImage *im, int *x, int *y, int decimate)
{
  int i,j;
  float sum = 0.0;
  float gamma=EDGE_DECAY;
  float theta=EDGE_THRESH;
  
  int width = im->width;
  int height = im->height;
  int got_one = BSSEG_FALSE;
  int lastx;
  int zoom = im->width / 10;
  
  for(j=zoom; j < height; j+=zoom)
    {
      got_one = BSSEG_FALSE;
      sum = 0.0;
      
      for(i=0; i < width && !got_one; i++)
	{
	  sum = gamma*sum + 
	    ( PixelSimple( im, i, j, decimate)
	     ?
	     1.0
	     :
	     0.0 );
	  if (sum >= theta)
	    {
	      *x = i;
	      *y = j;
	      return BSSEG_TRUE;
	    }
	}
    }
  return BSSEG_FALSE;
}





/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Raster
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


static int Raster(iveImage *im,
		 iveBitmap *outMap,
		 iveBitmap *inMap,
		 int zoom,
		 Point *centroid )
{
  int i,j;
  static iveBitmap *doneMap = NULL;
  register int width = im->width;
  register int height = im->height;
  if ( doneMap == NULL ) 
    doneMap = iveBitmapMake( height, width );
  
  for(i=1; i<(width-1); i++) 
    for(j=1;j<(height-1);j++) 
      {
	iveBitmapPutBit( outMap, j, i, 
			PixelClass(im, zoom, i, j )); 
	
	iveBitmapSetBit( doneMap, j, i );
      }

  return (1);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * ZoomGrow
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* do a multi-resolution grow */
static int ZoomGrow(iveImage *im,
		     iveBitmap *outMap,
		     iveBitmap *inMap,
		     int zoom,
		     Point *centroid )
{
  register int width = im->width;
  register int height = im->height;

  if (Grow(im,outMap,inMap,zoom*8,centroid))
    {
      Grow(im,outMap,inMap,zoom*4,centroid);
      Grow(im,outMap,inMap,zoom*2,centroid);
      Grow(im,outMap,inMap,zoom*1,centroid);
    }
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Grow
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void
pfMainGetBoundingBox( Point* left, Point* right )
{
  extern Point lowerLeft;
  extern Point lowerRight;
  left->x = lowerLeft.x;
  left->y = lowerLeft.y;
  right->x = lowerRight.x;
  right->y = lowerRight.y;
}
/* #def hack */
extern Point seed_points_for_morphology[1024];

#define NHOOD 1
/* use neighborhood information and do a grow */
static int Grow(iveImage *im,
		     iveBitmap *outMap,
		     iveBitmap *inMap,
		     int zoom,
		     Point *centroid )
{
  static Point stack1[MAX_POINTS_IN_STACK];
  static Point stack2[MAX_POINTS_IN_STACK];
  static Point neighbors[8];
  register Point *burning = stack1;
  register Point *catching = stack2;
  register Point *temp;
  register int burnptr = 0;
  register int catchptr = 0;
  int burn_timer = 0;
  int done = BSSEG_FALSE;
  register int width = im->width;
  register int height = im->height;
  int startx, starty;
  int found_blob = BSSEG_FALSE;
  int points = 0;
  int i;

  bzero(visitedbulk, sizeof(int)*width*height);


/* #define DEBUG_CENTROID */
#ifdef DEBUG_CENTROID
  fprintf(stderr,"start Centroid: %d %d\n", 
	 centroid->x, centroid->y);
#endif

  /* Light the first point on fire */
  if (!(centroid->x == 0 && centroid->y == 0) && 
      (PixelClass(im, 
		  zoom,
		  centroid->x,
		  centroid->y)))
    {
      /* snap to zoom grid */
      burning[burnptr].x = centroid->x - (centroid->x % zoom);
      burning[burnptr].y = centroid->y - (centroid->y % zoom);

      seed_points_for_morphology[burnptr] = burning[burnptr];
      burnptr++;
    }
  else
    {
      
      found_blob = GetEdgePoint(im, &startx, &starty, zoom);
      if (!found_blob)
	{
	  return (0);
	}
      
      burning[burnptr].x = startx - (startx % zoom) + zoom;
      burning[burnptr].y = starty - (starty % zoom) + zoom;
      seed_points_for_morphology[burnptr] = burning[burnptr];
      burnptr++;
    }



  {
    register int class;
    register int num_classes = bsClassifyGetNumClasses();
    float fx, fy;
    int x,y;

    for(class=0;class<num_classes;class++)
      {
	bsClassifyGetClassPosition(class, &fx, &fy);
	x = (int) fx;
	y = (int) fy;
	
	if(bsClassifyGetConfidence(class)>0.5 && 
	   PixelClass(im, zoom, x, y))
	  {
	    burning[burnptr].x = x - (x % zoom) + zoom;
	    burning[burnptr].y = y - (y % zoom) + zoom;
	    seed_points_for_morphology[burnptr] = burning[burnptr];
	    burnptr++;
	  }
      }
  }
	
  {
    register Point left,right;
    register int x,y;

    pfMainGetBoundingBox(&left, &right);

    y = left.y - 10;

    for(x=left.x;(x<right.x && !(PixelClass(im, zoom, x, y)));x++) ;
    if (PixelClass(im, zoom, x, y)) {
	  burning[burnptr].x = x - (x % zoom) + zoom;
	  burning[burnptr].y = y - (y % zoom) + zoom;
	  burnptr++;
	  
	  bsClassifySetClassPosition(pfRIGHTFOOT, (float) x, (float) y);
    }

    for(x=right.x;(x>left.x && !(PixelClass(im, zoom, x, y)));x--) ;
    if (PixelClass(im, zoom, x, y)) {
	  burning[burnptr].x = x - (x % zoom) + zoom;
	  burning[burnptr].y = y - (y % zoom) + zoom;
	  burnptr++;
	      
	  bsClassifySetClassPosition(pfLEFTFOOT, (float) x, (float) y);
    }

    for (x= centroid->x-10;x<=centroid->x+10; x+=10){
      for (y = 0; y < height; y+=5) {
	if (PixelClass(im, zoom, x, y)) {

	  burning[burnptr].x = x - (x % zoom) + zoom;
	  burning[burnptr].y = y - (y % zoom) + zoom;
	  seed_points_for_morphology[burnptr] = burning[burnptr];
	  burnptr++;
	}
      }
    }
  }
  seed_points_for_morphology[burnptr].x = 0;
  seed_points_for_morphology[burnptr].y = 0; /* marks end */
  centroid->x = 0;
  centroid->y = 0;
	
  while (!done)
    {
      
      while (burnptr--) /* burning stack loop */

	{
	  int dx, dy;
	  int x, y;
	  int count;
	  int class;
	  
	  /* Get next burning point */
	  x = burning[burnptr].x;
	  y = burning[burnptr].y;
	  count = 0;
	  /* Look at all neighbors, using subsample factor */

	  {
	    /* be nice to pipelines */
	    /* memoization will catch duplicate execution */
	    int w = im->width;
	    int h = im->height;
	    getonepix(im, zoom, w, h, x-zoom, y-zoom);
	    getonepix(im, zoom, w, h, x-zoom, y);
	    getonepix(im, zoom, w, h, x-zoom, y+zoom);
	    getonepix(im, zoom, w, h, x,      y+zoom);
	    getonepix(im, zoom, w, h, x,      y-zoom);
	    getonepix(im, zoom, w, h, x+zoom, y-zoom);
	    getonepix(im, zoom, w, h, x+zoom, y);
	    getonepix(im, zoom, w, h, x+zoom, y+zoom);
	  }
	  
	  for(dx = -NHOOD; dx <= NHOOD; dx++) /* neighbors */
	    for(dy = -NHOOD; dy <= NHOOD; dy++)
	      {
		int nx = x + (dx*zoom);
		int ny = y + (dy*zoom);
		
		if (nx < 1 || ny < 1 ||
		    nx >= (width-1) || ny >= (height-1))
		  continue; /* neighbors */	
		
		if (visited[nx][ny] )
		  continue; /* neighbors */
		
		/* If foreground pixel, add to catching */
		if (class = PixelClass(im, 
				       zoom,
				       nx,
				       ny))
		  {
		    count++;
		    catching[catchptr].x = nx;
		    catching[catchptr].y = ny;
		    catchptr++;
		    iveBitmapPutBit( outMap, ny, nx, class);
		    points++;
		    centroid->x += nx;
		    centroid->y += ny;
		  }
	     
	  
		visited[nx][ny] = 1;
	      } /* End neighbors loop */
	  
	} /* burning stack loop */
      
      /* swap burning and  non burning stacks */
      burnptr = catchptr;
      catchptr = 0;
      temp = catching;
      catching = burning;
      burning = temp;
      
      /* If no more burning, we are done */
      if (burnptr == 0) 
	done = BSSEG_TRUE;
      
      burn_timer++;
    } /* While (!done) */
#if DEBUG_THAD
  fprintf(stderr,"Points in blob: %d\n", points);
#endif

  if(points != 0)
    {
      centroid->x /= points;
      centroid->y /= points;
    }
  else
    {
      centroid->x = 0;
      centroid->y = 0;
    }

#ifdef DEBUG_CENTROID
  fprintf(stderr,"end Centroid: %d %d\n", 
	 centroid->x, centroid->y);
#endif

#if 0
  /* what really needs to be done here is return the largest
    of the various seed blobs */
  { int hackx = centroid->x; int hacky=im->height/2;
    if (! PixelClass(im, zoom, hackx, hacky)) {
      int k;
      for (k=0; k<64; k++)  {
	hackx = centroid->x - k;
	if (hackx < 0) hackx = 0;
	if (PixelClass(im, zoom, hackx, hacky)) {
	  centroid->x = hackx; centroid->y = hacky;
	  break;
	}
	hackx = centroid->x + k;
	if (hackx >= im->width ) hackx = im->width - 1;
	if (PixelClass(im, zoom, hackx, hacky)) {
	  centroid->x = hackx; centroid->y = hacky;
	  break;
	}
      }
      } else {
      	  centroid->x = hackx; centroid->y = hacky;
	}
  }
#endif
#ifdef DEBUG_CENTROID
  fprintf(stderr,"hacked Centroid: %d %d\n", 
	 centroid->x, centroid->y);
#endif


  return (1);
}



