#include <rpc/rpc.h>
#include <string.h>
#include <unistd.h>

#include <pfinder_clnt.h>

#ifdef __sgi
#include <bstring.h>
#else
#include <memory.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
void msleep(long msec)
{
  struct timeval timeout;
  long usec = msec * 1000;
  timeout.tv_sec = usec / 1000000;
  timeout.tv_usec = usec % 1000000; 
  select (0, NULL, NULL, NULL, &timeout);
}

static const int NFRAMES = 100;
static void howslow() 
{
  static int total = -1;
  static int count = 0;
  static struct timeval start;
  static struct timeval end;
  static double rate;
  static double time_sec;

  total++;
  // count = number of calls since init, *including* this one
  // Thusly, when count == NFRAMES, the time elapsed is for NFRAMES
  //  calls to processFrame()

  if (count==0)  // only happens once
     {
	gettimeofday(&start);
	count = 1;
	return;
     } 
  else if ( count++ == NFRAMES ) 
     {
	gettimeofday( &end );
	time_sec = (end.tv_usec - start.tv_usec)/1000000.0 + 
	   (end.tv_sec - start.tv_sec);
	rate = (double) NFRAMES / time_sec; // frames per second
	// fprintf ( stderr, "[%2.2f Hz]\n", rate );
	fprintf ( stderr, "[%2.2f Hz]\n", rate);
	start = end;
	count = 1;
	return;
     } 
  // putc('.',stderr);fflush(stderr);
}

int main (int argc, char ** argv) {
  int i=0;
  CLIENT *cl;
  pfperson *p;
  pfopts o;
  pfv2f *scr;
  bzero(&o,sizeof(pfopts));
  if (argc<2) exit(0);

  if (!( cl = clnt_create( argv[1], PFINDERPROG, PFINDERVERS, "tcp")))
    clnt_pcreateerror( argv[1] ), exit(1);

  o.kalman = 0;
  o.alive_coords = 1;
  o.f2 = o.g = o.c2 = o.c3 = NO;

  o.sync = NO;
  o.f3   = YES;
  o.cc2d = YES;
  o.cc3d = YES;

  if(argc>2)
    {
      o.cc2d = NO;
      if(argv[2][0] == 'f') o.f3 = YES;
      else if(argv[2][2] == '2') o.cc2d = YES;
      else if(argv[2][2] == '3') o.cc3d = YES;
    }

  scr = pfscrsize_1( NULL, cl );
  if (!scr) clnt_perror(cl, argv[1]), exit(-1);

  for (;;i++)
    {
      p = pf_1(&o, cl);
      if (!p) clnt_perror(cl, argv[1]), exit(-1);

      // printf("p frame %d (%f,%f)\n", p->n, scr->x, scr->y);
	  
      if (p->f3.opt) 
	{
	  pffeat3d_s *f3 = p->f3.pffeat3d_u.feats;
	  
// 	  printf("centroid (%7.2f %7.2f,%7.2f)\n",
// 		 f3->centroid.x, 
// 		 f3->centroid.y, 
// 		 f3->centroid.z );
	  
	  
	}
      if (p->cc3d.opt) 
	{
	  int j;
	  int num = p->cc3d.pfcc3d_u.cc.cc_len;
	  pfcc3d_s *cc = p->cc3d.pfcc3d_u.cc.cc_val;
	  for (j=0; j<num; j++)
	    {
// 	      printf(" %1d ", j);
// 	      printf("gm %7.2f %7.2f %7.2f ",
// 		     cc[j].gmean[0], cc[j].gmean[1], cc[j].gmean[2]);
//               printf("ge0 %6.2f %6.2f %6.2f %6.2f ", 
// 		     cc[j].gL[0], 
// 		     cc[j].gS[0].x, 
// 		     cc[j].gS[0].y, 
// 		     cc[j].gS[0].z);
//               printf("ge1 %6.2f %6.2f %6.2f %6.2f ", 
// 		     cc[j].gL[1], 
// 		     cc[j].gS[1].x, 
// 		     cc[j].gS[1].y, 
// 		     cc[j].gS[1].z);
//               printf("ge2 %6.2f %6.2f %6.2f %6.2f\n", 
// 		     cc[j].gL[2], 
// 		     cc[j].gS[2].x, 
// 		     cc[j].gS[2].y, 
// 		     cc[j].gS[2].z);

	    }
	}

      if (p->cc2d.opt) 
	{
	  int j;
	  int num = p->cc2d.pfcc2d_u.cc.cc_len;
	  pfcc2d_s *cc = p->cc2d.pfcc2d_u.cc.cc_val;
	  
	  for (j=0; j<num; j++)
	    {
// 	      printf(" %1d cm %6.2f %6.2f %6.2f ", j,
// 		     cc[j].cmean[0], cc[j].cmean[1], cc[j].cmean[2]);
// 	      printf("gm %6.2f %6.2f ",
// 		     cc[j].gmean[0], cc[j].gmean[1]);
//               printf("ge0 %6.2f %6.2f %6.2f ", 
// 		     cc[j].gL[0], cc[j].gS[0].x, cc[j].gS[0].y);
//               printf("ge1 %6.2f %6.2f %6.2f\n", 
// 		     cc[j].gL[1], cc[j].gS[1].x, cc[j].gS[1].y);

	    }
	}
      //msleep(100);    
      howslow();
    }
}


