/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

/* iveImage.h */
/* RCS: $Header: /v/projects/pfinder/distribution/src2/../../src/RCS/iveImage.h,v 9.0 1997/03/08 15:56:57 cwren Exp $ */

#include <stdio.h>

#ifndef _IVEIMAGE_H_
#define _IVEIMAGE_H_


#define DIGITIZEYUV 
/* #undef  DIGITIZEYUV */

#ifdef DIGITIZEYUV
#define iveImageBytesPerPixel (2)
#else
#define iveImageBytesPerPixel (4)
#endif

#define iveBPP iveImageBytesPerPixel

#define iveImageBytesPerWord (4)
#define iveBPW iveImageBytesPerWord

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * class ivePixel
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef DIGITIZEYUV

typedef struct
{
  unsigned char y;
  unsigned char u;
  unsigned char v;
} ivePixel;

typedef enum 
{ iveImageY=0, iveImageU=1, iveImageV=2 } ivePixelCh;

#define ivePixelRed( p ) ( (p)->y )
#define ivePixelGreen( p ) ( (p)->y )
#define ivePixelBlue( p ) ( (p)->y )
#define ivePixelChannel( p, ch ) ( *(((unsigned char *)(p)) + (ch)) )

#else /* DIGITIZEYUV */

typedef enum 
{ iveImageAlpha, iveImageBlue, iveImageGreen, iveImageRed } ivePixelCh;

typedef struct
{
  unsigned char alpha;
  unsigned char blue;
  unsigned char green;
  unsigned char red;
} ivePixel;

#define ivePixelChannel( p, ch ) ( *(((unsigned char *)(p)) + 3 - (ch)) )
#define ivePixelColor( p, c ) ( *(((unsigned char *)(p)) + (c)) )
#define ivePixelRed( p ) ( (p)->red )
#define ivePixelGreen( p ) ( (p)->green )
#define ivePixelBlue( p ) ( (p)->blue )

extern unsigned char ivePixelGetColor( ivePixel *, ivePixelCh );
extern unsigned char ivePixelGetRed( ivePixel * );
extern unsigned char ivePixelGetGreen( ivePixel * );
extern unsigned char ivePixelGetBlue( ivePixel * );

extern void ivePixelPutColor( ivePixel *, ivePixelCh, unsigned char );
extern void ivePixelPutRed( ivePixel *, unsigned char );
extern void ivePixelPutGreen( ivePixel *, unsigned char );
extern void ivePixelPutBlue( ivePixel *, unsigned char );

#endif /* DIGITIZEYUV */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * class iveImage
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

typedef struct
{
  int height;
  int width;
  int pixelCount;
  int byteCount;
  unsigned char *data;
} iveImage;

#ifdef DIGITIZEYUV

/* note iveImageColor is bad for abstraction; it should be done away with */
/* Instead, use iveImageGetPixel */

#define RawIndex( im, y, x ) (((y)*(im)->width+(x)))
#define PacketIndex( raw ) ( ((raw-raw%2)*2) )
#define YIndex( raw ) ( 2*raw + 1 )
#define UIndex( raw ) ( PacketIndex( raw ) )
#define VIndex( raw ) ( PacketIndex( raw ) + 2 )
#define iveImageIndexI(im,i,c) ((c==0?YIndex(i):(c==1?UIndex(i):VIndex(i))))
#define iveImageIndex( im, y, x, c ) (iveImageIndexI(im,RawIndex(im, y, x),c))
#define iveImageColor( im, y, x, c ) ((im)->data[iveImageIndex(im,y,x,c)])
#define iveImageColorI( im, i, c ) ( (im)->data[iveImageIndexI(im,i,c)] )

#else /* DIGITIZEYUV */

#define iveImageIndex( i, y, x, c ) (((y)*(i)->width+(x))*iveBPW + (c))
#define iveImageColor( i, y, x, c ) ((i)->data[iveImageIndex(i,y,x,c)])
#define iveImageColorI( im, i, c ) ( (im)->data[(i) * iveBPW + (c)] )

#endif /* DIGITIZEYUV */

extern iveImage * iveImageMake( int height, int width );
extern iveImage * iveImageCopy( iveImage *src, iveImage *dst );
extern iveImage * iveImageMakeCopy( iveImage *im );
extern void       iveImageFree( iveImage * );
extern iveImage * iveImageInit( iveImage *im, 
				int height,
				int width, 
				unsigned char * );
extern int iveImageReadRaw( iveImage *, FILE *stream );
extern int iveImageWriteRaw( iveImage *, FILE *stream );
extern ivePixel *iveImageGetPixel( iveImage *, int y, int x, ivePixel * );
extern ivePixel *iveImageGetPixelI( iveImage *, int i, ivePixel * );
extern void iveImagePutPixel( iveImage *, int y, int x, ivePixel * );

#if 0
/* not used *//* only work for RGB right now! */
extern iveImage * iveImageMakeLoad( char *file );
extern int iveImageWrite( iveImage *, char *file );
extern void iveImageGetRGB( iveImage *, int y, int x, int *, int *, int * );
extern void iveImagePutRGB( iveImage *, int y, int x, int, int, int );
extern unsigned char iveImageGetColor(iveImage *, int y, int x, ivePixelCh);
extern void iveImagePutColor( iveImage *,
			      int y,
			      int x,
			      ivePixelCh, 
			      unsigned char );

#endif

#endif /* _IVEIMAGE_H_ */
