/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

// lite.c++

#define Time XTIME
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <Inventor/SoDB.h>
#undef Time
#include "lite.h++"

extern "C"
{
#include "bsClassify.h"
#include "bsXclient.h"
  extern void setxdisplay( Display *, XtAppContext );
  extern void bsXHandleEvents(void);
};
extern int UPDATING_CONTOUR;	

pfCamera* lite::camera = 0;
pfFeatures lite::features;
pfPersonSize lite::personSize( 0. );
pfGesture lite::gestures;
_feature_data lite::kalmanFeatures;
float lite::mmtNoise = 1.;
float lite::accelNoise = 1.;
int lite::kfGrain = 3;
int lite::personExists = 0;
int lite::reportFreq = 10000;
int lite::reportMove = 0;
// int lite::handChooseContour = 1.95;
// int lite::footChooseContour = 2.;
pfSemaphore lite::kalmanSemaphore;
int lite::contourGroove = 0;
int lite::printGestures = 0;

static
inline
void
error( char *fn, char *msg, int doexit )
{
    cerr << "lite:: " << fn << ": " << msg << endl;
    if( doexit )
       exit( 1 );
}

void
lite::_reconcileColorClass( pfClassDesig pfclass,
		      float* cont3dFeat,
		      boolean& contValid,
		      float centroid[2],
		      float boundingbox[2] )
{
  // Chooses to trust EITHER the contour feature OR the color feature
  // if CONTOUR, then override the color feature in bsClassify
  // if COLOR, then override feature in lite

  // added a buffer zone of blending

      
  int height = bsClassifyYdim();
  int width  = bsClassifyXdim();
  
  float trustColor_confidence = .5;    // [0,1]
  float trustContour_confidence = 1.5; // stddev
  int   Contour_direction = 1; // y

  if ((pfclass==pfRIGHTHAND) || (pfclass==pfLEFTHAND))
    {
      trustContour_confidence = 1.0;
      Contour_direction = 0; // x
    }
  
  // valid bit seems to always be set if person exists
  // and this code only happens if person exists
  // so this is a little redundant, but maybe that 
  // will change.  I wish whever computes the contour 
  // feature would give me a score, but since they don't
  // I'll just wing it. 

  if (contValid == pfTRUE)
    {
      float cont2dFeat_screen[3]; // find a common rep
      float cont2dFeat_pixels[2];
      lite::Project( cont3dFeat, cont2dFeat_screen );
      cont2dFeat_pixels[0] = cont2dFeat_screen[0] * width;
      cont2dFeat_pixels[1] = cont2dFeat_screen[1] * height;
      
      float color2dFeat_pixels[2];
      bsClassifyGetClassPosition(pfclass, 
				 &color2dFeat_pixels[0], 
				 &color2dFeat_pixels[1]);
      float color2dFeat_screen[2];
      color2dFeat_screen[0] = color2dFeat_pixels[0] / width;
      color2dFeat_screen[1] = color2dFeat_pixels[1] / height;
      
      // calculate useful metrics

      float Contour_Score = fabs((cont2dFeat_pixels[Contour_direction] - 
				  centroid[Contour_direction]) / 
				 boundingbox[Contour_direction]);

      if ((pfclass==pfRIGHTHAND) || (pfclass==pfLEFTHAND))
	{
	  if (cont2dFeat_pixels[1] < (centroid[1] - boundingbox[1]))
	    Contour_Score = 0.0;
	}

      float Color_score = bsClassifyGetConfidence( pfclass );
      float Distance = sqrt(((cont2dFeat_pixels[0]-color2dFeat_pixels[0]) *
			     (cont2dFeat_pixels[0]-color2dFeat_pixels[0])) + 
			    ((cont2dFeat_pixels[1]-color2dFeat_pixels[1]) *
			     (cont2dFeat_pixels[1]-color2dFeat_pixels[1])));
      

      // who do we trust? 
      if (Contour_Score > trustContour_confidence) 
	{
	  // assume we have valid contour features
	  if (Color_score > trustColor_confidence)
	    {
	      // some support for a color feature

	      if ((Distance / boundingbox[Contour_direction]) > 0.2)
		{
		  // color strayed from a good feature
		  // give the color class a clue
		  
		  bsClassifySetClassPosition( pfclass, 
					     cont2dFeat_pixels[0], 
					     cont2dFeat_pixels[1]);
		}
	      else
		{
		  // everyone is happy... just leave them be
		}
	    }
	  else
	    {
	      // all we have is contour feature
	      // give the color class a clue
	      
	      bsClassifySetClassPosition( pfclass, 
					 cont2dFeat_pixels[0], 
					 cont2dFeat_pixels[1]);
	    }
	    
	}
      else
	{
	  // assume we have invalid contour features
	  if (Color_score > trustColor_confidence)
	    {
	      // some support for a color feature
	      // run with it
	      float color3dFeat[3];
	      lite::BackProject( color2dFeat_screen, color3dFeat );
	      cont3dFeat[0] = color3dFeat[0];
	      cont3dFeat[1] = color3dFeat[1];
	      cont3dFeat[2] = color3dFeat[2];
	    }
	  else
	    {
	      // we got nothin!
	      // but something is better than nothing!

	      bsClassifySetClassPosition( pfclass, 
					 cont2dFeat_pixels[0], 
					 cont2dFeat_pixels[1]);
	      //contValid = pfFALSE;
	    }
	}
    }
  return;
}

void
lite::_initializeColorClasses()
{
  float centroid_pixels[2];
  float stdev_pixels[2];
  
  bsClassifyGetClassPosition(pfGLOBAL, 
			     &centroid_pixels[0], 
			     &centroid_pixels[1]);
  bsClassifyGetClassBB(pfGLOBAL, 
		       &stdev_pixels[0], 
		       &stdev_pixels[1] );
  
  bsClassifyInitForegroundPrior( pfCHEST );
  bsClassifySetClassPosition( pfCHEST, 
			     centroid_pixels[0],
			     centroid_pixels[1] -
			     0.9 * stdev_pixels[1]);
  
  bsClassifyInitForegroundPrior( pfWAIST);
  bsClassifySetClassPosition( pfWAIST, 
			     centroid_pixels[0],
			     centroid_pixels[1]);
  
  bsClassifyInitForegroundPrior( pfTHIGHS );
  bsClassifySetClassPosition( pfTHIGHS, 
			     centroid_pixels[0],
			     centroid_pixels[1] +
			     0.9 * stdev_pixels[1]);
}

void
lite::_reconcileWithColorClasses()
{
    float centroid_pixels[2];
    float stdev_pixels[2];

    bsClassifyGetClassPosition(pfGLOBAL, 
			       &centroid_pixels[0], 
			       &centroid_pixels[1]);
    bsClassifyGetClassBB(pfGLOBAL, 
			 &stdev_pixels[0], 
			 &stdev_pixels[1] );

    stdev_pixels[1] = personSize.value();
    stdev_pixels[0] = 0.5 * stdev_pixels[1];
      

    _feature_data& f = lite::getFeatures().getCurrent();
    
    _reconcileColorClass( pfHEAD,
			  f.userplaneHead,
			  f.headValid,
			  centroid_pixels,
			  stdev_pixels);
    _reconcileColorClass( pfLEFTHAND,
			  f.userplaneLeftHand,
			  f.leftHandValid,
			  centroid_pixels,
			  stdev_pixels);
    _reconcileColorClass( pfRIGHTHAND,
			  f.userplaneRightHand,
			  f.rightHandValid,
			  centroid_pixels,
			  stdev_pixels);
    _reconcileColorClass( pfLEFTFOOT,
			  f.userplaneLeftFoot,
			  f.leftFootValid,
			  centroid_pixels,
			  stdev_pixels);
    _reconcileColorClass( pfRIGHTFOOT,
			  f.userplaneRightFoot,
			  f.rightFootValid,
			  centroid_pixels,
			  stdev_pixels);
}

void
lite::initialize( void )
{
  SoDB::init();
  if( !( camera = new pfCamera( "dogCameraParameters" ) ) )
     error( "initialize", "couldn't construct pfCamera", 1 );
}

static
void
_draw_features( _feature_data& f )
{
    const int which = 0;
    const int width = 160;
    const int height = 120;
    // int x = f.imageHead[0] * width;
    // int y = f.imageHead[1] * height;
    // cerr << x << ", " << y << ":  " << f.imageHead[0] << ", " << f.imageHead[1] << endl;

    if( f.headValid )
      bsXDrawColorCross( f.imageHead[0]*width, f.imageHead[1]*height,
			15, 4, which );
    if( f.leftHandValid )
      bsXDrawColorCross( f.imageLeftHand[0]*width, f.imageLeftHand[1]*height,
			15, 3, which );
    if( f.rightHandValid )
      bsXDrawColorCross( f.imageRightHand[0]*width, f.imageRightHand[1]*height,
			15, 2, which );
    if( f.leftFootValid )
      bsXDrawColorCross( f.imageLeftFoot[0]*width, f.imageLeftFoot[1]*height, 
			15, 5, which );
    if( f.rightFootValid )
      bsXDrawColorCross( f.imageRightFoot[0]*width, f.imageRightFoot[1]*height,
			15, 7, which );
}

void
lite::interpretPfContour( pfContour& cont )
{
  static int visits = 0;
  static int personExisted = 0;	// did personExist last time?

  personExists = features.ProjectPfContourToUserPlane( cont, camera );
  if( personExists )
    {
      personExisted++;

      features.Find( camera );
      if( !contourGroove )
	_reconcileWithColorClasses();
      if( personExisted < 20 ) 
	_initializeColorClasses();
    }
    else
      {
	features.SetNotValid();
	personExisted = 0;	
      }

  features.swapBuffers();

  kalmanUpdate( features.getLatest() );

  if( personExists && gestures.bodyMoving() )
    personSize.standingHeadSample( kalmanFeatures.userplaneHead[1] );

  features.getLatest().userheight = personSize.value();
  
  if( personExists )
    gestures.Update( kalmanFeatures.userplaneHead,
		     kalmanFeatures.userplaneLeftHand,
		     kalmanFeatures.userplaneRightHand,
		     kalmanFeatures.userplaneCentroid,
		     personSize.value() );
  else 
    gestures.Clear();

  if( !( visits % reportFreq ) )
    {
      _draw_features( features.getLatest() );
      cerr << ">HEIGHT: " << personSize.value() << endl;
      if( printGestures )
	gestures.Print( cerr );
    }

  visits++;
}

_feature_data&
lite::kalmanUpdate( _feature_data& in )
{
    kalmanSemaphore.localLOCK();

    _feature_data& out = kalmanFeatures;

    vector3Copy( in.userplaneHead, out.userplaneHead );
    vector3Copy( in.userplaneLeftHand, out.userplaneLeftHand );
    vector3Copy( in.userplaneRightHand, out.userplaneRightHand );
    vector3Copy( in.userplaneLeftFoot, out.userplaneLeftFoot );
    vector3Copy( in.userplaneRightFoot, out.userplaneRightFoot );
    vector3Copy( in.userplaneCentroid, out.userplaneCentroid );

    kalmanSemaphore.localUNLOCK();

    camera -> projectToScreen2( out.userplaneHead, out.imageHead );
    camera -> projectToScreen2( out.userplaneLeftHand, out.imageLeftHand );
    camera -> projectToScreen2( out.userplaneRightHand, out.imageRightHand );
    camera -> projectToScreen2( out.userplaneLeftFoot, out.imageLeftFoot );
    camera -> projectToScreen2( out.userplaneRightFoot, out.imageRightFoot );
    camera -> projectToScreen2( out.userplaneCentroid, out.imageCentroid );

    return out;
}


void lite::dumpContour( FILE *stream )
{
    int numpoints = features.getLatest().numContourPoints;
    float *array = features.getLatest().imageContour;
    fwrite( &numpoints, sizeof(int), 1, stream );
    fwrite( array, sizeof(float), 2 * numpoints, stream );
}


void lite::dumpFeatures( FILE *stream )
{
  fprintf(stream, "%f %f %f %f %f %f\n",
	  features.getLatest().imageHead[0],
	  features.getLatest().imageHead[1],
	  features.getLatest().imageLeftHand[0],
	  features.getLatest().imageLeftHand[1],
	  features.getLatest().imageRightHand[0],
	  features.getLatest().imageRightHand[1]);
}


