/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

#include <stream.h>
#include <rpc/rpc.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <signal.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <bstring.h>

#include "pfinder_svc.h"

#define MAXREQ 10

struct svc_req REQUESTS[MAXREQ];
struct pfopts OPTS[MAXREQ];
int SYNCH=0;
int UPDATING_CONTOUR=0;
int REQ_ARR_LOCK=0;
int num_requests=0;

void cache_pf_request(pfopts *opts, struct svc_req *req) {
  while (REQ_ARR_LOCK) sginap(CLK_TCK/100+1);
  if (num_requests == MAXREQ-1) {
    fputs("threaded req buf overflow!\n",stderr);
  } else {
#ifdef DEBUG_RPC_CACHE
    fprintf(stderr,"caching req %d : %d %d \n",num_requests,
	    opts->sync, opts->g);
#endif
    bcopy(opts,OPTS + num_requests,sizeof(pfopts));
    bcopy(req,REQUESTS + num_requests,sizeof(svc_req));
    num_requests++;
  }
}


void call_cached_pf_reqests(void) {

  REQ_ARR_LOCK=1;SYNCH=1;

  while (num_requests) {
    num_requests--;
    pfopts *popt = OPTS + num_requests;
    struct svc_req *preq = REQUESTS + num_requests;

#ifdef DEBUG_RPC_CACHE
    fprintf(stderr,"execing req %d : %d %d\n",num_requests,popt->sync,popt->g);
#endif

    popt->sync = 0;

#ifdef DEBUG_RPC_CACHE
    fprintf(stderr,"calling req %d : %d %d\n",num_requests,popt->sync,popt->g);
#endif

    // while (UPDATING_CONTOUR) sginap(CLK_TCK/100+1);

    if (preq->rq_vers == PFINDERVERS)
      {
	pfperson *r = pf_1(popt,preq);
	if (svc_sendreply(preq->rq_xprt,  (xdrproc_t) xdr_pfperson, r)) {
	  svcerr_systemerr(preq->rq_xprt);
	}
      }
    else if (preq->rq_vers == PFINDERVERS2)
      {
	pfinfo *r = pf_2(popt,preq);
	if (svc_sendreply(preq->rq_xprt,  (xdrproc_t) xdr_pfinfo, r)) {
	  svcerr_systemerr(preq->rq_xprt);
	}
      }
    else
      {
	fprintf(stderr,"pf-cache: unrecognized version: %d\n",
		preq->rq_vers);
      }	
  }
  REQ_ARR_LOCK=0;SYNCH=0;
}


