/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

// pfCamera.h++

#ifndef _pfCamera_H_
#define _pfCamera_H_

#include <stream.h>
#include <fstream.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/SbPList.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/projectors/SbPlaneProjector.h>

class BackProject {
    SbMatrix spm_static;
    SbMatrix spm1;
    SoTransform *xf;
    SbVec3f cp;
    SbMatrix matrix;
    SbMatrix m2;
    SbPlane plane;
    SbMatrix spm;
    SbMatrix proj;
    float planeZ;
    void setProjectionPlane(float z);
public:
    BackProject( SoPerspectiveCamera *cam );
    void backProjectPt( const SbVec3f &sp1, SbVec3f &projPt, float z );
};

class pfCamera
{
private:
    SoPerspectiveCamera* camera;
    BackProject *projector;
    float heightToWidth;
public:
    pfCamera( char *filename );

    // [0,1], y-down	==>	3-D
    void projectScreenYtoWorldY( const float in[2], float out[3], float z );
    int projectScreenYtoWorldY( int n, float* in, float* out, float z );
    void projectYtoZFloor( const float in[2], float out[3] );

    // 3-D		==>	[0,1], y-down
    void projectToScreen( const float in[3], float out[3] );
    void projectToScreen2( const float in[3], float out[2] );
    float AspectRatio() { return heightToWidth; }
};

#endif /* _pfCamera_H_ */
