/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

// pfSemaphore.h++

#ifndef _pfSemaphore_H_
#define _pfSemaphore_H_

#include <pthread.h>

class pfSemaphore
{
  pthread_mutex_t the_mutex;
  
public:
  pfSemaphore()
  {
    if (pthread_mutex_init(&the_mutex, (pthread_mutexattr_t *) NULL))
      cerr << "pfSemaphore::pfSemaphore() error creating mutex" << endl;
  };

  ~pfSemaphore()
  {
    Lock();
    pthread_mutex_destroy(&the_mutex);
  };

  inline void Lock() {
    pthread_mutex_lock(&the_mutex);
  };
  
  inline void Unlock() { 
    pthread_mutex_unlock(&the_mutex);
  };

  // There are for compatibility with old interface
  inline void localLOCK() { Lock(); };
  inline void remoteLOCK() { Lock(); };

  inline void localUNLOCK() { Unlock(); };
  inline void remoteUNLOCK() { Unlock(); };

};

#endif /* _pfSemaphore_H_ */
