#include <bstring.h>
#include <stdlib.h>
#include <iostream.h>
#include "pfTextureMap.h++"

extern "C" void bsColorToRgb(float *mean);

pfTextureMap::pfTextureMap(int user_max_width,
			   int user_max_height)
{
  if ((user_max_width == 0) || (user_max_height == 0))
    {
      cerr << "pfTextureMap::pfTextureMap: width and height must be non-zero" << endl;
      exit(1);
    }
  
  data = new unsigned char[4 * user_max_width * user_max_height];
  max_width = user_max_width;
  max_height = user_max_height;
}

pfTextureMap::~pfTextureMap()
{
  delete[] data;
}

void
pfTextureMap::copyFromIveImage(iveImage *image,
			       int min_x, 
			       int min_y,
			       int max_x,
			       int max_y)
{
  if ((image->width > max_width) ||
      (image->height > max_height))
    {
      cerr << "pfTextureMap::copyFromIveImage: incoming image is bigger than my buffer!" << endl;
      return;
    }

  if ((min_x > max_x) || (min_y > max_y))
    {
      cerr << "pfTextureMap::copyFromIveImage: min and max reversed" << endl;
      return;
    }

  int bytes_copied = 0;

  semaphore.localLOCK();

#ifdef DIGITIZEYUV
  float tmp[3];

  for (int y = min_y; y <= max_y; y++)
    {
      for (int x = min_x; x <= max_x; x++)
	{
	  tmp[0] = (float) iveImageColor( image, y, x, iveImageY );
	  tmp[1] = (float) iveImageColor( image, y, x, iveImageU );
	  tmp[2] = (float) iveImageColor( image, y, x, iveImageV );

	  bsColorToRgb(tmp);

	  /* output RBGA */
	  data[bytes_copied++] = (unsigned char) tmp[0];
	  data[bytes_copied++] = (unsigned char) tmp[1];
	  data[bytes_copied++] = (unsigned char) tmp[2];
	  data[bytes_copied++] = 255;
	}
    }
#else
  for (int y = min_y; y <= max_y; y++)
    {
      for (int x = min_x; x <= max_x; x++)
	{
	  /* output RBGA */
	  data[bytes_copied++] = iveImageColor( image, y, x, iveImageRed );
	  data[bytes_copied++] = iveImageColor( image, y, x, iveImageGreen );
	  data[bytes_copied++] = iveImageColor( image, y, x, iveImageBlue );
	  data[bytes_copied++] = iveImageColor( image, y, x, iveImageAlpha );
	}
    }
#endif  
  
  width = max_x - min_x + 1;
  height = max_y - min_y + 1;

  semaphore.localUNLOCK();
}

int
pfTextureMap::getMaxTextureWidth()
{
  return max_width;
}

int
pfTextureMap::getMaxTextureHeight()
{
  return max_height;
}

int
pfTextureMap::getTextureWidth()
{
  return width;
}

int
pfTextureMap::getTextureHeight()
{
  return height;
}

const unsigned char *
pfTextureMap::getTextureData()
{
  return data;
}
