/* 
RTRACK region tracker software
MIT Media Lab Vision Group

All source, documentation, and examples in this distribution are
Copyright 1995 MIT. Distribution and use for non-commercial research
and development is permitted. All rights reserved.

Comments/questions to trevor@media.mit.edu
*/
/* rtrack-main.c++ */

/* RCS: $Header: /v/projects/pfinder/distribution/src2/../../src/RCS/rtrack-main.c++,v 9.0 1997/03/08 15:56:57 cwren Exp $ */

#include <stdio.h>
#include <stdlib.h>
#define _BSD_SIGNALS
#include <signal.h>
#include <unistd.h>
#include <iostream.h>
#include <bstring.h>
#include <math.h>

#include "Parse.h"
#include "VLImage.h++"

extern "C"
{
#include "ivePoint.h"
#include "iveBitmap.h"
#include "iveImage.h"
#include "bsXclient.h"
#include "bsColor.h"
#include "bsClassify.h"
#include "bsSegment.h"
}

static int readMode = 0;	// read input from stdin
static int pauseMode = 0;	// loop on current image
static int dumpMode = 0;	// dump input to dumpstream
static int expert = 0;		// show all the gory sliders
static int dumpCount = 0;
static int dumpDelay = 0;
static char *chroma = NULL;
static char *dumpFile = NULL;
static FILE *dumpstream = stdout;
static int decimate = 1;

static int useXInterface = 1;
static int verbose = 0;
static int numberOfBGFrames = 40;
static iveImage **BackgroundFrames = NULL;
static iveImage **imageBuffers = NULL;
Point lowerLeft;
Point lowerRight;

int RPCEXITCMD = 0;

int INITCLASSFLAG = 0;

static int HOWSLOW = 1;
int FRAMEZERO = 0;

int SEGMENT=1; /* continous */
int SEGMENT_TRIGGER=0; /* one shot */

void cleanup( void );
void cleanup_sig( ... );

extern void startRpcServer(void);
extern void stopRpcServer(void);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * howslow
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static const int NFRAMES = 25;

static void howslow() 
{
  static int total = -1;
  static int count = 0;
  static struct timeval start;
  static struct timeval end;
  static double rate;
  static double time_sec;

  total++;
  // count = number of calls since init, *including* this one
  // Thusly, when count == NFRAMES, the time elapsed is for NFRAMES
  //  calls to processFrame()

  if (count==0)  // only happens once
     {
	gettimeofday(&start);
	count = 1;
	return;
     } 
  else if ( count++ == NFRAMES ) 
     {
	gettimeofday( &end );
	time_sec = (end.tv_usec - start.tv_usec)/1000000.0 + 
	   (end.tv_sec - start.tv_sec);
	rate = (double) NFRAMES / time_sec; // frames per second
	// fprintf ( stderr, "[%2.2f Hz]\n", rate );
	fprintf ( stderr, "[%2.2f Hz, (%d frames, %d total, %.6f sec) %d]\n", 
		  rate, NFRAMES, (int)(30.*time_sec), time_sec, total );
	start = end;
	count = 1;
	return;
     } 
  // putc('.',stderr);fflush(stderr);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * image flow control
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int STEP = 1;

void bsMainSetPause(int newvalue) {
  pauseMode = newvalue;
}

int bsMainGetPause() {
  return pauseMode;
}

void bsMainStep()
{
  STEP =1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * processFrame
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void
processFrame( iveImage *image)
{
  /* does not need frameNumber as an arg; can keep track internally */

  static int frameNumber = -1;
  static Point centroid;
  static iveBitmap *newMap = (iveBitmap *)NULL;
  static iveBitmap *oldMap = (iveBitmap *)NULL;
  iveBitmap *tempMap = NULL;
#ifdef DONOTHING
  howslow(); return;
  /* NOTREACHED */
#endif
  frameNumber++;

  if (FRAMEZERO) {
    frameNumber = 0;
    FRAMEZERO = 0;
  }  

  if( newMap == (iveBitmap *)NULL )
    {
      int height = image->height;
      int width = image->width;
      
      newMap = iveBitmapMake( height, width );
      oldMap = iveBitmapMake( height, width );
      iveBitmapZero( oldMap );
      centroid.x = 0;
      centroid.y = 0;	
    }

  if( dumpMode )
    {
      if( dumpCount>0 && frameNumber >= dumpCount )
	{
	  if( imageBuffers )
	    {
	      for( int i = 0; i < dumpCount; i++ )
		iveImageWriteRaw( imageBuffers[i], dumpstream );
	    }
	  cleanup();
	}
      if( verbose )
	cerr << "pfinder: dumping frame " << frameNumber << endl;
      if( useXInterface )
	{
	  bsXDisplayImage( image, 1);
	}
      if( imageBuffers )
	iveImageCopy( image, imageBuffers[frameNumber] );
      else
	iveImageWriteRaw( image, dumpstream );


      return;
    }
  

  if (frameNumber < numberOfBGFrames) {
    
    BackgroundFrames[frameNumber] = iveImageMakeCopy(image);
    
    if( useXInterface ) {
      bsXDisplayImage( image , 1);
    }
    
    return;
  }

  if (frameNumber == numberOfBGFrames) 
    {
      if (!chroma) 
	{
	  bsClassifyMakeBackgroundClasses(BackgroundFrames, numberOfBGFrames);
	} else {
	  bsClassifyChromakey(chroma);
	}
    }
  
  /* classify */

  if( HOWSLOW ) howslow();
  if( RPCEXITCMD ) cleanup();

  int dosegment = SEGMENT || SEGMENT_TRIGGER;
  SEGMENT_TRIGGER = 0;
  if (dosegment) {
      int RASTERPAINT = 1;
      
      if( useXInterface  ) 
	{
	  bsXNewCanvas( 1 );  

	  if (!RASTERPAINT)
	    bsXCommitPaint( 1 );

	}

      iveBitmapZero( newMap );
      bsSegmentFindSilhouette( image, newMap, oldMap, 
			      decimate, &(centroid));

      
  
      if( useXInterface  ) 
	{
	  if (RASTERPAINT)
	    bsXDisplayBitmap( newMap , 1);
	  else
	    bsXCommitPaint( 1 ); 
	  
	}
  } else {
	  if( useXInterface )
	    bsXDisplayImage(image, 1);

  }


  if( useXInterface ) {
	bsXSync();	
  }



  tempMap = oldMap;
  oldMap = newMap;
  newMap = tempMap;
}

 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * cleanup
 * cleanup_sig
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void cleanup_do( void ) {
      
      stopRpcServer();

      if ( useXInterface ) {
//	    cerr << getpid() << " cleaning up X" << endl;
	    bsXClose();
      }

      if (!readMode) {
//	    cerr << getpid() << " cleaning up VL" << endl;
	    VLImage::Close();
      }
}

void cleanup( void )
{
  cleanup_do();

  exit(1);
}

void cleanup_sig( ... )
{
  cerr << "Caught Signal" << getpid()  << endl;       
  cleanup();
}

static void 	FrameZero() { FRAMEZERO = 1; }

static int  bsMainGetSegment() { return !SEGMENT; }
static void bsMainSetSegment(int newvalue) { SEGMENT = !newvalue; }

void verboseCallback( void * )
{
  bsColorSetVerbose( 1 );
  bsClassifySetVerbose( 1 );
  VLImage::Verbose( 1 );
}


void chromaModeCallback( void * )
{
  bsSegmentSetThresh(10.0);
}

void dumpModeCallback( void * )
{
  dumpMode = 1;
  readMode = 0;
}

void readModeCallback( void * )
{
  readMode = 1;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * main
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


void
help( void *progname )
{
  Parse::VerboseUsage( 1 );
  Parse::Usage( (char *) progname );
  cleanup();
}

int 
main(int argc, char *argv[])
{
  int zoom = 6;

  int width = 640;
  int height = 480;
  int nbuffers = 3;
  int input_connector = VLImage::INDY_INDYCAM;
  static char *inputHelp = "\n"\
     "\t\t\t-1: INDY_INDYCAM\n"\
     "\t\t\t0: INDY_COMPOSITE\n"\
     "\t\t\t1: INDY_SVIDEO\n"\
     "\t\t\t0: GALILEO_SVIDEO_1\n"\
     "\t\t\t1: GALILEO_SVIDEO_2\n"\
     "\t\t\t2: GALILEO_SVIDEO_3\n"\
     "\t\t\t3: GALILEO_COMPOSITE_1\n"\
     "\t\t\t4: GALILEO_COMPOSITE_2\n"\
     "\t\t\t5: GALILEO_COMPOSITE_3\n"\
     "\t\t\t6: GALILEO_YRB_1\n"\
     "\t\t\t7: GALILEO_YRB_2\n"\
     "\t\t\t\t";

  ParseRuleFlag( 'v', &verbose, 1, "Main program verbose mode on" );
  ParseRuleFlag( 'v', verboseCallback, 0, "bsColor and VL modules verbose" );
  ParseRuleFlag( 'v', &expert, 1, "show all controls" );
  ParseRuleFlag( 'X', &useXInterface, !useXInterface, "Use X Interface" );
  ParseRuleTaggedArgInt( "b", &numberOfBGFrames, "Number of BG Frames" );
  ParseRuleTaggedArgInt( "d", &decimate, "Decimation factor" );
  ParseRuleTag( "D", dumpModeCallback, 0, 
		"Dump mode on, images from VL to cout" );
  ParseRuleTaggedArgInt( "D", &dumpCount, "Number of frames to dump" );
  ParseRuleTag( "R", readModeCallback, 0, "Read mode on, images fm cin" );
  ParseRuleTag( "single", &pauseMode, !pauseMode, 
	       "loop on first captured image");
  ParseRuleTaggedArgInt( "w", &width, "Width of images" );
  ParseRuleTaggedArgInt( "h", &height, "Height of images" );
  ParseRuleTaggedArgInt( "size", &height, "Height of images", "height" );
  ParseRuleTaggedArgInt( "size", &width, "Width of images", "width" );
  ParseRuleTaggedArgInt( "z", &zoom, "Zoom factor" );
  ParseRuleTaggedArgInt( "in", &input_connector, inputHelp, "conn" );
  ParseRuleTaggedArgInt( "nb", &nbuffers, "Number of VL buffers to use", "number" );

  ParseRuleTag( "noslow", &HOWSLOW, 0, "Turn off rate reporting" );
  ParseRuleTag( "help", help, argv[0], "Prints usage" );
  
  // rules for getting file names for dumping output to files other than stdout

  ParseRuleTaggedArg("dumpfile", &dumpFile, "File to dump input to");

  ParseRuleTag( "chroma", chromaModeCallback, 0, 
		"use chromakey instead of aquiring background" );
  ParseRuleTaggedArg("chroma", &chroma, "chroma to key on %2x%2x%2x");

  if( Parse::ParseCmdLine( argc, argv ) )
    {
      Parse::VerboseUsage( 0 );
      Parse::Usage( argv[0] );
      Parse::VerboseParse();
      Parse::ParseCmdLine( argc, argv );
      cleanup();
    }
  
  signal(SIGINT, cleanup_sig);
  signal(SIGPIPE, cleanup_sig); 
  signal(SIGHUP, cleanup_sig);
  signal(SIGILL, cleanup_sig);

  if (chroma)
    {
      numberOfBGFrames = 0;
    }

  if (dumpFile)
    {
      if (verbose)
	cerr << "dumpFile was " << dumpFile << endl;
      if ((dumpstream = fopen(dumpFile, "w")) == (FILE *) NULL)
	{
	  cerr << "Error opening dump file" << endl;
	  cleanup();
	}
    }

  if( !readMode )
    {
      VLImage::NBuffers( nbuffers );
#ifdef DIGITIZEYUV
      VLImage::Encoding( VLImage::YUV );
#else
      VLImage::Encoding( VLImage::RGB );
#endif
      VLImage::Zoom( zoom );
      VLImage::Input( input_connector );
      VLImage::Init();
      height = VLImage::GetHeight();
      width = VLImage::GetWidth();
    }
  if( dumpMode )
    {
      if( verbose )
	cerr << "pfinder: VL frame size: height= " << height
	     << " width= " << width << endl;
      if( dumpMode && dumpCount <= 100 && dumpCount>0)
	{
	  imageBuffers = (iveImage **) malloc( dumpCount*sizeof(iveImage*) );
	  if( imageBuffers )
	    {
	      for( int i = 0; i < dumpCount; i++ )
		{
		  imageBuffers[i] = iveImageMake( height, width );
		  if( ! imageBuffers[i] )
		    {
		      fprintf( stderr, "pfinder: failed to malloc image %d\n", i );
		      exit( 1 );
		    }
		}
	    }
	}
    }
  

  if( useXInterface )
    {
      bsXInitialize(width, height, 
		    1,
		    0, NULL );
      
      if ( !dumpMode ) 
	{
	  bsXMakeSliderProc( bsSegmentSetThresh, "FG/BG Threshhold",
			    bsSegmentGetThreshMax(), 0.0, 
			    bsSegmentGetThresh() );
	  if (expert) {
		bsXMakeSliderProc( bsColorSetNormalFloor, "Normalize Floor",
				   1.0, 0.0, 
				   bsColorGetNormalFloor() );
	  }
	}
      
      if ( !dumpMode ) 
	{
	  if (expert)
	    {
	      bsXMakeButtonProc(bsMainSetSegment, "Pass-Through",
				bsMainGetSegment() );
	    }

	  if (expert)
	    {
	      bsXMakeSeparator();
	      bsXMakeButtonProc(bsSegmentSetZoomGrow, "Grow",
				bsSegmentGetZoomGrow() );
	      bsXMakeButtonProc(bsSegmentSetSmooth, "Smooth",
				bsSegmentGetSmooth() );
	    }
	  
	  if (expert) 
	    {
	      bsXMakeSeparator();
	      bsXMakeButtonProc(bsColorSetNormalize, "Normalize Enable",
				bsColorGetNormalize() );
	    }

	  bsXMakeSeparator();
	  if (expert)
	    {
	      bsXMakeButtonProc( bsMainSetPause, "Pause",
				bsMainGetPause());
	      bsXMakePushButton( bsMainStep, "Step" );
	      bsXMakeSeparator();
	    }
	  bsXMakePushButton( FrameZero, "Reaquire BG" );
	  
	}
      
      bsXMakePushButton( cleanup, "panic: core dump" );
    }

  if( !dumpMode )
     {
       bsClassifyInitialize(width, height, 1);
       
       BackgroundFrames = (iveImage **) malloc( sizeof(iveImage *) * 
					       numberOfBGFrames );
       bsClassifyNewForegroundClass( );
     }
  
  startRpcServer();

  if( readMode )		// read from a file
    {
      iveImage *image = iveImageMake( height, width ); 
      
      for(;;)
	 {
	   if( useXInterface ) 
	     bsXHandleEvents();
	   
	   if (pauseMode && !STEP)
	     {
	       processFrame( image );
	     }
	   else
	     {
	       if (iveImageReadRaw( image, stdin ))
		 {
		   fprintf(stderr,"end of input\n");
		   cleanup();
		 }
	       processFrame( image );
	     }
	   STEP = 0;
	 }
    }
  else				// live or dumpMode
    {
      sleep( dumpDelay );

#ifdef DIGITIZEYUV
      VLYuvImage picture;
#else
      VLRgbImage picture;
#endif
      iveImage wrapper;
      for(;;)
	 {
	   if( useXInterface )
	      bsXHandleEvents();
      
	   if (pauseMode && !STEP)
	     {
	       processFrame( &wrapper );
	     }
	   else
	     {
	       picture.Grab();
	       
	       iveImageInit(&wrapper, 
			    picture.Height(), 
			    picture.Width(), 
			    picture );
	       processFrame( &wrapper );
	       
	       picture.Free();
	     }

	   STEP = 0;

	 }
    }      
}



